#  coding=utf-8
#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""
Common environment vars / settings
"""
import logging
from typing import List, Optional
from urllib.parse import urlparse

from pydantic import BaseSettings, Field, validator


class ApiSettings(BaseSettings):
    """
    Main api settings. The pydantic BaseSettings class makes
    accessible environment variables by setting attributes.

    See <https://pydantic-docs.helpmanual.io/usage/settings/>

    only_bulk_api: (ONLY_BULK_API env var)
         If True, activate only bulk and search endpoints

    elasticseach: (ELASTICSEARCH env var)
        The elasticsearch endpoint for datasets persistence

    cors_origins: (CORS_ORIGINS env var)
        List of host patterns for CORS origin access

    docs_enabled: True
        If True, enable openapi docs endpoint at /api/docs

    es_records_index_shards:
        Configures the number of shards for dataset records index creation. Default=1

    es_records_index_replicas:
        Configures the number of shard replicas for dataset records index creation. Default=0

    disable_es_index_template_creation: (DISABLE_ES_INDEX_TEMPLATE_CREATION env var)
         Allowing advanced users to create their own es index settings and mappings. Default=False

    """

    __LOGGER__ = logging.getLogger(__name__)

    __DATASETS_INDEX_NAME__ = "ar.datasets"
    __DATASETS_RECORDS_INDEX_NAME__ = "ar.dataset.{}"

    elasticsearch: str = "http://localhost:9200"
    elasticsearch_ssl_verify: bool = True
    elasticsearch_ca_path: Optional[str] = None
    cors_origins: List[str] = ["*"]

    docs_enabled: bool = True

    namespace: str = Field(default=None, regex=r"^[a-z]+$")

    enable_migration: bool = Field(
        default=False,
        description="If enabled, try to migrate data from old rubrix installation",
    )

    # Analyzer configuration
    default_es_search_analyzer: str = "standard"
    exact_es_search_analyzer: str = "whitespace"
    # This line will be enabled once words field won't be used anymore
    # wordcloud_es_search_analyzer: str = "multilingual_stop_analyzer"

    es_records_index_shards: int = 1
    es_records_index_replicas: int = 0

    metadata_fields_limit: int = Field(
        default=50, gt=0, le=100, description="Max number of fields in metadata"
    )

    enable_telemetry: bool = True

    telemetry_key: Optional[str] = None

    @property
    def dataset_index_name(self) -> str:
        ns = self.namespace
        if ns:
            return f"{self.namespace}.{self.__DATASETS_INDEX_NAME__}"
        return self.__DATASETS_INDEX_NAME__

    @property
    def dataset_records_index_name(self) -> str:
        ns = self.namespace
        if ns:
            return f"{self.namespace}.{self.__DATASETS_RECORDS_INDEX_NAME__}"
        return self.__DATASETS_RECORDS_INDEX_NAME__

    @property
    def old_dataset_index_name(self) -> str:
        index_name = ".rubrix<NAMESPACE>.datasets-v0"
        ns = self.namespace
        if ns is None:
            return index_name.replace("<NAMESPACE>", "")
        return index_name.replace("<NAMESPACE>", f".{ns}")

    @property
    def old_dataset_records_index_name(self) -> str:
        index_name = ".rubrix<NAMESPACE>.dataset.{}.records-v0"
        ns = self.namespace
        if ns is None:
            return index_name.replace("<NAMESPACE>", "")
        return index_name.replace("<NAMESPACE>", f".{ns}")

    def obfuscated_elasticsearch(self) -> str:
        """Returns configured elasticsearch url obfuscating the provided password, if any"""
        parsed = urlparse(self.elasticsearch)
        if parsed.password:
            return self.elasticsearch.replace(parsed.password, "XXXX")
        return self.elasticsearch

    class Config:
        # TODO: include a common prefix for all argilla env vars.
        env_prefix = "ARGILLA_"
        fields = {
            # TODO(@frascuchon): Remove in 0.20.0
            "elasticsearch": {
                "env": ["ELASTICSEARCH", f"{env_prefix}ELASTICSEARCH"],
            },
            "elasticsearch_ssl_verify": {
                "env": [
                    "ELASTICSEARCH_SSL_VERIFY",
                    f"{env_prefix}ELASTICSEARCH_SSL_VERIFY",
                ]
            },
            "cors_origins": {"env": ["CORS_ORIGINS", f"{env_prefix}CORS_ORIGINS"]},
            "docs_enabled": {"env": ["DOCS_ENABLED", f"{env_prefix}DOCS_ENABLED"]},
            "es_records_index_shards": {
                "env": [
                    "ES_RECORDS_INDEX_SHARDS",
                    f"{env_prefix}ES_RECORDS_INDEX_SHARDS",
                ]
            },
            "es_records_index_replicas": {
                "env": [
                    "ES_RECORDS_INDEX_REPLICAS",
                    f"{env_prefix}ES_RECORDS_INDEX_SHARDS",
                ]
            },
        }


settings = ApiSettings()



#  coding=utf-8
#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""
Dataset models definition
"""

from typing import Any, Dict, Optional

from pydantic import BaseModel, Field

from argilla._constants import DATASET_NAME_REGEX_PATTERN
from argilla.server.commons.models import TaskType
from argilla.server.services.datasets import ServiceBaseDataset


class UpdateDatasetRequest(BaseModel):
    """
    Modifiable fields for dataset

    Attributes:
    -----------
    tags:
        Dataset tags used for better organize or classify information
    metadata:
        Extra information that could be interested to include
    """

    tags: Dict[str, str] = Field(default_factory=dict)
    metadata: Dict[str, Any] = Field(default_factory=dict)


class _BaseDatasetRequest(UpdateDatasetRequest):
    name: str = Field(regex=DATASET_NAME_REGEX_PATTERN, description="The dataset name")


class CreateDatasetRequest(_BaseDatasetRequest):
    task: TaskType = Field(description="The dataset task")


class CopyDatasetRequest(_BaseDatasetRequest):
    """
    Request body for copy dataset operation
    """

    target_workspace: Optional[str] = None


class Dataset(_BaseDatasetRequest, ServiceBaseDataset):
    """
    Low level dataset data model

    Attributes:
    -----------
    task:
        The dataset task type. Deprecated
    owner:
        The dataset owner
    created_at:
        The dataset creation date
    last_updated:
        The last modification date
    """

    task: TaskType



#  coding=utf-8
#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import itertools
from typing import Iterable, Optional

from fastapi import APIRouter, Depends, Query, Security
from fastapi.responses import StreamingResponse

from argilla.server.apis.v0.handlers import (
    metrics,
    token_classification_dataset_settings,
)
from argilla.server.apis.v0.helpers import deprecate_endpoint
from argilla.server.apis.v0.models.commons.model import BulkResponse
from argilla.server.apis.v0.models.commons.params import (
    CommonTaskHandlerDependencies,
    RequestPagination,
)
from argilla.server.apis.v0.models.token_classification import (
    TokenClassificationAggregations,
    TokenClassificationBulkRequest,
    TokenClassificationDataset,
    TokenClassificationQuery,
    TokenClassificationRecord,
    TokenClassificationSearchRequest,
    TokenClassificationSearchResults,
)
from argilla.server.apis.v0.validators.token_classification import DatasetValidator
from argilla.server.commons.config import TasksFactory
from argilla.server.commons.models import TaskType
from argilla.server.errors import EntityNotFoundError
from argilla.server.helpers import takeuntil
from argilla.server.responses import StreamingResponseWithErrorHandling
from argilla.server.security import auth
from argilla.server.security.model import User
from argilla.server.services.datasets import DatasetsService
from argilla.server.services.tasks.token_classification import (
    TokenClassificationService,
)
from argilla.server.services.tasks.token_classification.metrics import (
    TokenClassificationMetrics,
)
from argilla.server.services.tasks.token_classification.model import (
    ServiceTokenClassificationQuery,
    ServiceTokenClassificationRecord,
)


def configure_router():
    task_type = TaskType.token_classification
    base_endpoint = f"/{{name}}/{task_type}"
    new_base_endpoint = f"/{task_type}/{{name}}"

    TasksFactory.register_task(
        task_type=task_type,
        dataset_class=TokenClassificationDataset,
        query_request=TokenClassificationQuery,
        record_class=ServiceTokenClassificationRecord,
        metrics=TokenClassificationMetrics,
    )

    router = APIRouter(tags=[task_type], prefix="/datasets")

    @router.post(
        path=f"{base_endpoint}:bulk",
        operation_id="bulk_records",
        response_model=BulkResponse,
        response_model_exclude_none=True,
    )
    async def bulk_records(
        name: str,
        bulk: TokenClassificationBulkRequest,
        common_params: CommonTaskHandlerDependencies = Depends(),
        service: TokenClassificationService = Depends(
            TokenClassificationService.get_instance
        ),
        datasets: DatasetsService = Depends(DatasetsService.get_instance),
        validator: DatasetValidator = Depends(DatasetValidator.get_instance),
        current_user: User = Security(auth.get_user, scopes=[]),
    ) -> BulkResponse:

        task = task_type
        owner = current_user.check_workspace(common_params.workspace)
        try:
            dataset = datasets.find_by_name(
                current_user,
                name=name,
                task=task,
                workspace=owner,
                as_dataset_class=TasksFactory.get_task_dataset(task_type),
            )
            datasets.update(
                user=current_user,
                dataset=dataset,
                tags=bulk.tags,
                metadata=bulk.metadata,
            )
        except EntityNotFoundError:
            dataset_class = TasksFactory.get_task_dataset(task)
            dataset = dataset_class.parse_obj({**bulk.dict(), "name": name})
            dataset.owner = owner
            datasets.create_dataset(user=current_user, dataset=dataset)

        records = [ServiceTokenClassificationRecord.parse_obj(r) for r in bulk.records]
        # TODO(@frascuchon): validator can be applied in service layer
        await validator.validate_dataset_records(
            user=current_user,
            dataset=dataset,
            records=records,
        )

        result = await service.add_records(
            dataset=dataset,
            records=records,
        )
        return BulkResponse(
            dataset=name,
            processed=result.processed,
            failed=result.failed,
        )

    @router.post(
        path=f"{base_endpoint}:search",
        response_model=TokenClassificationSearchResults,
        response_model_exclude_none=True,
        operation_id="search_records",
    )
    def search_records(
        name: str,
        search: TokenClassificationSearchRequest = None,
        common_params: CommonTaskHandlerDependencies = Depends(),
        include_metrics: bool = Query(
            False, description="If enabled, return related record metrics"
        ),
        pagination: RequestPagination = Depends(),
        service: TokenClassificationService = Depends(
            TokenClassificationService.get_instance
        ),
        datasets: DatasetsService = Depends(DatasetsService.get_instance),
        current_user: User = Security(auth.get_user, scopes=[]),
    ) -> TokenClassificationSearchResults:

        search = search or TokenClassificationSearchRequest()
        query = search.query or TokenClassificationQuery()

        dataset = datasets.find_by_name(
            user=current_user,
            name=name,
            task=task_type,
            workspace=common_params.workspace,
            as_dataset_class=TasksFactory.get_task_dataset(task_type),
        )
        results = service.search(
            dataset=dataset,
            query=ServiceTokenClassificationQuery.parse_obj(query),
            sort_by=search.sort,
            record_from=pagination.from_,
            size=pagination.limit,
            exclude_metrics=not include_metrics,
        )

        return TokenClassificationSearchResults(
            total=results.total,
            records=[TokenClassificationRecord.parse_obj(r) for r in results.records],
            aggregations=TokenClassificationAggregations.parse_obj(results.metrics)
            if results.metrics
            else None,
        )

    def scan_data_response(
        data_stream: Iterable[TokenClassificationRecord],
        chunk_size: int = 1000,
        limit: Optional[int] = None,
    ) -> StreamingResponseWithErrorHandling:
        """Generate an textual stream data response for a dataset scan"""

        async def stream_generator(stream):
            """Converts dataset scan into a text stream"""

            def grouper(n, iterable, fillvalue=None):
                args = [iter(iterable)] * n
                return itertools.zip_longest(fillvalue=fillvalue, *args)

            if limit:
                stream = takeuntil(stream, limit=limit)

            for batch in grouper(
                n=chunk_size,
                iterable=stream,
            ):
                filtered_records = filter(lambda r: r is not None, batch)
                yield "\n".join(
                    map(
                        lambda r: r.json(by_alias=True, exclude_none=True),
                        filtered_records,
                    )
                ) + "\n"

        return StreamingResponseWithErrorHandling(
            stream_generator(data_stream), media_type="application/json"
        )

    @router.post(
        path=f"{base_endpoint}/data",
        operation_id="stream_data",
    )
    async def stream_data(
        name: str,
        query: Optional[TokenClassificationQuery] = None,
        common_params: CommonTaskHandlerDependencies = Depends(),
        limit: Optional[int] = Query(None, description="Limit loaded records", gt=0),
        service: TokenClassificationService = Depends(
            TokenClassificationService.get_instance
        ),
        datasets: DatasetsService = Depends(DatasetsService.get_instance),
        current_user: User = Security(auth.get_user, scopes=[]),
        id_from: Optional[str] = None,
    ) -> StreamingResponse:
        """
            Creates a data stream over dataset records

        Parameters
        ----------
        name
            The dataset name
        query:
            The stream data query
        common_params:
            Common query params
        limit:
            The load number of records limit. Optional
        service:
            The dataset records service
        datasets:
            The datasets service
        current_user:
            Request user
        id_from:
            If provided, read the samples after this record ID

        """
        query = query or TokenClassificationQuery()
        dataset = datasets.find_by_name(
            user=current_user,
            name=name,
            task=task_type,
            workspace=common_params.workspace,
            as_dataset_class=TasksFactory.get_task_dataset(task_type),
        )
        data_stream = map(
            TokenClassificationRecord.parse_obj,
            service.read_dataset(
                dataset=dataset,
                query=ServiceTokenClassificationQuery.parse_obj(query),
                id_from=id_from,
                limit=limit,
            ),
        )

        return scan_data_response(
            data_stream=data_stream,
            limit=limit,
        )

    token_classification_dataset_settings.configure_router(router)
    metrics.configure_router(
        router,
        cfg=TasksFactory.get_task_by_task_type(task_type),
    )

    return router


router = configure_router()



#  coding=utf-8
#  coding=utf-8
#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import json
from typing import Any, Dict, List, Optional, Type, TypeVar, Union

import httpx

from argilla.client.sdk.client import AuthenticatedClient
from argilla.client.sdk.commons.errors import GenericApiError
from argilla.client.sdk.commons.errors_handler import handle_response_error
from argilla.client.sdk.commons.models import (
    BulkResponse,
    ErrorMessage,
    HTTPValidationError,
    Response,
)
from argilla.client.sdk.text2text.models import Text2TextBulkData
from argilla.client.sdk.text_classification.models import TextClassificationBulkData
from argilla.client.sdk.token_classification.models import TokenClassificationBulkData

_TASK_TO_ENDPOINT = {
    TextClassificationBulkData: "TextClassification",
    TokenClassificationBulkData: "TokenClassification",
    Text2TextBulkData: "Text2Text",
}


def build_param_dict(
    id_from: Optional[str], limit: Optional[int]
) -> Optional[Dict[str, Union[str, int]]]:
    params = {}
    if id_from:
        params["id_from"] = id_from
    if limit:
        params["limit"] = limit
    return params


def bulk(
    client: AuthenticatedClient,
    name: str,
    json_body: Union[
        TextClassificationBulkData, TokenClassificationBulkData, Text2TextBulkData
    ],
) -> Response[BulkResponse]:
    url = f"{client.base_url}/api/datasets/{name}/{_TASK_TO_ENDPOINT[type(json_body)]}:bulk"

    response = httpx.post(
        url=url,
        headers=client.get_headers(),
        cookies=client.get_cookies(),
        timeout=client.get_timeout(),
        json=json_body.dict(by_alias=True),
    )

    return build_bulk_response(response, name=name, body=json_body)


async def async_bulk(
    client: AuthenticatedClient,
    name: str,
    json_body: Union[
        TextClassificationBulkData, TokenClassificationBulkData, Text2TextBulkData
    ],
) -> Response[BulkResponse]:
    url = f"{client.base_url}/api/datasets/{name}/{_TASK_TO_ENDPOINT[type(json_body)]}:bulk"

    async with httpx.AsyncClient() as async_client:
        response = await async_client.post(
            url=url,
            headers=client.get_headers(),
            cookies=client.get_cookies(),
            timeout=client.get_timeout(),
            json=json_body.dict(by_alias=True),
        )

    return build_bulk_response(response, name=name, body=json_body)


def build_bulk_response(
    response: httpx.Response, name: str, body: Any
) -> Response[BulkResponse]:

    if 200 <= response.status_code < 400:
        return Response(
            status_code=response.status_code,
            content=response.content,
            headers=response.headers,
            parsed=BulkResponse(**response.json()),
        )

    return handle_response_error(response, name=name, body=body)


T = TypeVar("T")


def build_data_response(
    response: httpx.Response, data_type: Type[T]
) -> Response[List[T]]:
    if 200 <= response.status_code < 400:
        parsed_responses = []
        for r in response.iter_lines():
            parsed_record = json.loads(r)
            try:
                parsed_response = data_type(**parsed_record)
            except Exception as err:
                raise GenericApiError(**parsed_record) from None
            parsed_responses.append(parsed_response)
        return Response(
            status_code=response.status_code,
            content=b"",
            headers=response.headers,
            parsed=parsed_responses,
        )

    content = next(response.iter_lines())
    data = json.loads(content)
    return handle_response_error(response, **data, parse_response=False)


def build_list_response(
    response: httpx.Response,
    item_class: Type[T],
) -> Response[Union[List[T], HTTPValidationError, ErrorMessage]]:
    parsed_response = response.json()

    if 200 <= response.status_code < 400:
        parsed_response = [item_class(**r) for r in parsed_response]
        return Response(
            status_code=response.status_code,
            content=response.content,
            headers=response.headers,
            parsed=parsed_response,
        )

    return handle_response_error(response, item_class=item_class)



#  coding=utf-8
#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import re
import warnings
from typing import Any, Dict, Iterable, List, Optional, Tuple

from opensearchpy import NotFoundError, OpenSearch, OpenSearchException, RequestError
from opensearchpy.exceptions import OpenSearchWarning
from opensearchpy.helpers import bulk as es_bulk
from opensearchpy.helpers import reindex, scan

from argilla.logging import LoggingMixin
from argilla.server.commons.models import TaskType
from argilla.server.daos.backend import query_helpers
from argilla.server.daos.backend.mappings.datasets import (
    DATASETS_INDEX_NAME,
    datasets_index_mappings,
)
from argilla.server.daos.backend.mappings.helpers import (
    mappings,
    tasks_common_mappings,
    tasks_common_settings,
)
from argilla.server.daos.backend.mappings.text2text import text2text_mappings
from argilla.server.daos.backend.mappings.text_classification import (
    text_classification_mappings,
)
from argilla.server.daos.backend.mappings.token_classification import (
    token_classification_mappings,
)
from argilla.server.daos.backend.metrics import ALL_METRICS
from argilla.server.daos.backend.metrics.base import ElasticsearchMetric
from argilla.server.daos.backend.search.model import (
    BackendRecordsQuery,
    BaseDatasetsQuery,
    SortableField,
    SortConfig,
)
from argilla.server.daos.backend.search.query_builder import EsQueryBuilder
from argilla.server.errors import EntityNotFoundError, InvalidTextSearchError
from argilla.server.errors.task_errors import MetadataLimitExceededError

try:
    import ujson as json
except ModuleNotFoundError:
    import json

from argilla.server.settings import settings


class ClosedIndexError(Exception):
    pass


class IndexNotFoundError(Exception):
    pass


class GenericSearchError(Exception):
    def __init__(self, origin_error: Exception):
        self.origin_error = origin_error


class backend_error_handler:
    def __init__(self, index: str):
        # Maybe a backend to detect the backend nature...
        self._index = index

    def __enter__(self):
        # This line disable all open search client warnings
        warnings.filterwarnings("ignore", category=OpenSearchWarning)
        pass

    def __exit__(self, exception_type, exception_value, traceback):
        if not exception_value:
            return
        try:
            raise exception_value from exception_value
        except RequestError as ex:
            detail = self.__get_es_error_detail__(ex, exception_value)
            if ex.error == "search_phase_execution_exception":
                detail = detail or ex.info["error"]
                raise InvalidTextSearchError(detail)
            elif ex.error == "index_closed_exception":
                raise ClosedIndexError(self._index)
            raise GenericSearchError(exception_value) from exception_value
        except NotFoundError as ex:
            raise IndexNotFoundError(ex)
        except OpenSearchException as ex:
            raise GenericSearchError(ex)

    def __get_es_error_detail__(self, ex, exception_value) -> Optional[str]:
        detail = exception_value.info["error"]
        detail = detail.get("root_cause")

        return detail[0].get("reason")


class ElasticsearchBackend(LoggingMixin):
    """
    Encapsulates logic about the communication, queries and index mapping
    transformations between DAOs layer and the elasticsearch backend.
    """

    _INSTANCE = None

    __HIGHLIGHT_PRE_TAG__ = "<@@-ar-key>"
    __HIGHLIGHT_POST_TAG__ = "</@@-ar-key>"
    __HIGHLIGHT_VALUES_REGEX__ = re.compile(
        rf"{__HIGHLIGHT_PRE_TAG__}(.+?){__HIGHLIGHT_POST_TAG__}"
    )

    __HIGHLIGHT_PHRASE_PRE_PARSER_REGEX__ = re.compile(
        rf"{__HIGHLIGHT_POST_TAG__}\s+{__HIGHLIGHT_PRE_TAG__}"
    )

    # TODO(@frascuchon): Once id is included as keyword in datasets index, we can discard this
    __MAX_NUMBER_OF_LISTED_DATASETS__ = 2500

    @classmethod
    def get_instance(cls) -> "ElasticsearchBackend":
        """
        Creates an instance of ElasticsearchBackend.

        This function is used in fastapi for resolve component dependencies.

        See <https://fastapi.tiangolo.com/tutorial/dependencies/>

        Returns
        -------

        """

        if cls._INSTANCE is None:
            es_client = OpenSearch(
                hosts=settings.elasticsearch,
                verify_certs=settings.elasticsearch_ssl_verify,
                ca_certs=settings.elasticsearch_ca_path,
                # Extra args to es configuration -> TODO: extensible by settings
                retry_on_timeout=True,
                max_retries=5,
            )
            cls._INSTANCE = cls(
                es_client,
                query_builder=EsQueryBuilder(),
                metrics={**ALL_METRICS},
                mappings={
                    TaskType.text_classification: text_classification_mappings(),
                    TaskType.token_classification: token_classification_mappings(),
                    TaskType.text2text: text2text_mappings(),
                },
            )

        return cls._INSTANCE

    def __init__(
        self,
        es_client: OpenSearch,
        query_builder: EsQueryBuilder,
        metrics: Dict[str, ElasticsearchMetric] = None,
        mappings: Dict[str, Dict[str, Any]] = None,
    ):
        self.__client__ = es_client
        self.__query_builder__ = query_builder
        self.__defined_metrics__ = metrics or {}
        self.__tasks_mappings__ = mappings

    @property
    def client(self):
        """The elasticsearch client"""
        return self.__client__

    @property
    def query_builder(self):
        """The query builder"""
        return self.__query_builder__

    def _list_documents(
        self,
        index: str,
        query: Dict[str, Any] = None,
        sort_cfg: Optional[List[Dict[str, Any]]] = None,
        size: Optional[int] = None,
        fetch_once: bool = False,
    ) -> Iterable[Dict[str, Any]]:
        """
        List ALL documents of an elasticsearch index
        Parameters
        ----------
        index:
            The index name
        sor_id:
            The sort id configuration
        query:
            The es query for filter results. Default: None
        sort_cfg:
            Customized configuration for sort-by id
        size:
            Amount of samples to retrieve per iteration, 1000 by default
        fetch_once:
            If enabled, will return only the `size` first records found. Default to: ``False``

        Returns
        -------
        A sequence of documents resulting from applying the query on the index

        """
        size = size or 1000
        query = query.copy() or {}
        if sort_cfg:
            query["sort"] = sort_cfg
        query["track_total_hits"] = False  # Speedup pagination
        response = self.__client__.search(index=index, body=query, size=size)
        while response["hits"]["hits"]:
            for hit in response["hits"]["hits"]:
                yield hit
            if fetch_once:
                break

            last_id = hit["_id"]
            query["search_after"] = [last_id]
            response = self.__client__.search(index=index, body=query, size=size)

    def _index_exists(self, index: str) -> bool:
        """
        Checks if provided index exists

        Parameters
        ----------
        index:
            The index name

        Returns
        -------
            True if index exists. False otherwise
        """
        return self.__client__.indices.exists(index)

    def _search(
        self,
        index: str,
        routing: str = None,
        size: int = 100,
        query: Dict[str, Any] = None,
    ) -> Dict[str, Any]:
        """
        Apply a search over an index.
        See <https://www.elastic.co/guide/en/elasticsearch/reference/current/search-search.html>

        Parameters
        ----------
        index:
            The index name
        routing:
            The routing key. Optional
        size:
            Number of results to return. Default=100
        query:
            The elasticsearch query. Optional

        Returns
        -------

        """
        with backend_error_handler(index=index):
            return self.__client__.search(
                index=index,
                body=query or {},
                routing=routing,
                track_total_hits=True,
                rest_total_hits_as_int=True,
                size=size,
            )

    def _create_index(
        self,
        index: str,
        force_recreate: bool = False,
        settings: Dict[str, Any] = None,
        mappings: Dict[str, Any] = None,
    ):
        """
        Applies a index creation with provided mapping configuration.

        See <https://www.elastic.co/guide/en/elasticsearch/reference/current/indices-create-index.html>

        Parameters
        ----------
        force_recreate:
            If True, the index will be recreated (if exists). Default=False
        settings:
            The index settings configuration
        mappings:
            The mapping configuration. Optional.

        """
        with backend_error_handler(index):
            if force_recreate:
                self._delete_index(index)
            if not self._index_exists(index):
                self.__client__.indices.create(
                    index=index,
                    body={"settings": settings or {}, "mappings": mappings or {}},
                    ignore=400,
                )

    def _create_index_template(
        self, name: str, template: Dict[str, Any], force_recreate: bool = False
    ):
        """
        Applies a index template creation with provided template definition.

        Parameters
        ----------
        name:
            The template index name
        template:
            The template definition
        force_recreate:
            If True, the template will be recreated (if exists). Default=False

        """
        with backend_error_handler(index=""):
            if force_recreate or not self.__client__.indices.exists_template(name):
                self.__client__.indices.put_template(name=name, body=template)

    def delete_index_template(self, index_template: str):
        """Deletes an index template"""
        with backend_error_handler(index=""):

            if self.__client__.indices.exists_index_template(index_template):
                self.__client__.indices.delete_template(
                    name=index_template, ignore=[400, 404]
                )

    def _delete_index(self, index: str, raises_error: bool = False):
        """Deletes an elasticsearch index"""
        with backend_error_handler(index=index):
            if self._index_exists(index):
                ignore_errors = [400, 404]
                if raises_error:
                    ignore_errors = []
                self.__client__.indices.delete(index, ignore=ignore_errors)

    def _add_document(self, index: str, doc_id: str, document: Dict[str, Any]):
        """
        Creates/updates a document in an index

        See <http://www.elastic.co/guide/en/elasticsearch/reference/current/docs-index_.html>

        Parameters
        ----------
        index:
            The index name
        doc_id:
            The document id
        document:
            The document source

        """
        with backend_error_handler(index=index):
            self.__client__.index(
                index=index, body=document, id=doc_id, refresh="wait_for"
            )

    def _get_document_by_id(self, index: str, doc_id: str) -> Optional[Dict[str, Any]]:
        """
        Get a document by its id

        See <https://www.elastic.co/guide/en/elasticsearch/reference/current/docs-get.html>

        Parameters
        ----------
        index:
            The index name
        doc_id:
            The document id

        Returns
        -------
            The elasticsearch document if found, None otherwise
        """

        try:
            if self.__client__.exists(index=index, id=doc_id):
                return self.__client__.get(index=index, id=doc_id)
        except NotFoundError:
            return None
        except Exception as ex:
            with backend_error_handler(index=index):
                raise ex

    def _delete_document(self, index: str, doc_id: str):
        """
        Deletes a document from an index.

        See <http://www.elastic.co/guide/en/elasticsearch/reference/current/docs-delete.html>

        Parameters
        ----------
        index:
            The index name
        doc_id:
            The document id

        Returns
        -------

        """
        with backend_error_handler(index=index):
            if self.__client__.exists(index=index, id=doc_id):
                self.__client__.delete(index=index, id=doc_id, refresh=True)

    def add_documents(
        self,
        id: str,
        documents: List[Dict[str, Any]],
    ) -> int:
        index = dataset_records_index(id)

        def doc_id(r):
            return r.get("id")

        def map_doc_2_action(doc: Dict[str, Any]) -> Dict[str, Any]:
            """Configures bulk action"""
            data = {
                "_op_type": "index",
                "_index": index,
                "_routing": None,  # TODO(@frascuchon): Use a sharding routing
                **doc,
            }

            id_ = doc_id(doc)
            if id_ is not None:
                data["_id"] = id_

            return data

        with backend_error_handler(index=index):
            success, failed = es_bulk(
                self.__client__,
                index=index,
                actions=map(map_doc_2_action, documents),
                raise_on_error=True,
                refresh="wait_for",
            )
            return len(failed)

    def _get_field_mapping(
        self,
        index: str,
        field_name: Optional[str] = None,
        exclude_subfields: bool = False,
    ) -> Dict[str, str]:
        """
            Returns the mapping for a given field name (can be as wildcard notation). The result
        consist on a dictionary with full field name as key and its type as value

        See <http://www.elastic.co/guide/en/elasticsearch/reference/7.13/indices-get-field-mapping.html>

        Parameters
        ----------
        index:
            The index name
        field_name:
            The field name pattern
        exclude_subfields:
            If True, exclude extra subfields from mappings definition

        Returns
        -------
            A dictionary with full field name as key and its type as value
        """
        try:
            response = self.__client__.indices.get_field_mapping(
                fields=field_name or "*",
                index=index,
                ignore_unavailable=False,
            )
            schema = self._extract_index_schema_from_response(
                index=index, es_response=response
            )
            data = {
                key: list(definition["mapping"].values())[0]["type"]
                for key, definition in schema["mappings"].items()
            }

            if exclude_subfields:
                # Remove `text`, `exact` and `wordcloud` fields
                def is_subfield(key: str):
                    for suffix in ["exact", "text", "wordcloud"]:
                        if suffix in key:
                            return True
                    return False

                data = {
                    key: value for key, value in data.items() if not is_subfield(key)
                }

            return data
        except NotFoundError:
            # No mapping data
            return {}
        except Exception as ex:
            with backend_error_handler(index=index):
                raise ex

    def _update_document(
        self,
        index: str,
        doc_id: str,
        document: Optional[Dict[str, Any]] = None,
        script: Optional[str] = None,
        partial_update: bool = False,
    ):
        """
        Updates a document in a given index

        Parameters
        ----------
        index:
            The index name
        doc_id:
            The document id
        document:
            The document data. Could be partial document info
        partial_update:
            If True, document contains partial info, and will be
            merged with stored document. If false, the stored document
            will be overwritten. Default=False

        Returns
        -------

        """
        # TODO: validate either doc or script are provided
        with backend_error_handler(index=index):
            if not partial_update:
                return self.__client__.index(
                    index=index, id=doc_id, body=document, refresh=True
                )

            body = {"script": script} if script else {"doc": document}
            return self.__client__.update(
                index=index,
                id=doc_id,
                body=body,
                refresh=True,
                retry_on_conflict=500,  # TODO: configurable
            )

    def open_index(self, index: str):
        """
        Open an elasticsearch index. If index is already open, this operation will do nothing

        See `<https://www.elastic.co/guide/en/elasticsearch/reference/7.11/indices-open-close.html>`_

        Parameters
        ----------
        index:
            The index name
        """
        with backend_error_handler(index=index):
            self.__client__.indices.open(
                index=index, wait_for_active_shards=settings.es_records_index_shards
            )

    def _close_index(self, index: str):
        """
        Closes an elasticsearch index. If index is already closed, this operation will do nothing.

        See `<https://www.elastic.co/guide/en/elasticsearch/reference/7.11/indices-open-close.html>`_

        Parameters
        ----------
        index:
            The index name
        """
        with backend_error_handler(index=index):
            self.__client__.indices.close(
                index=index,
                ignore_unavailable=True,
                wait_for_active_shards=settings.es_records_index_shards,
            )

    def _clone_index(self, index: str, clone_to: str, override: bool = True):
        """
        Clone an existing index. During index clone, source must be setup as read-only index. Then, changes can be
        applied

        See `<https://www.elastic.co/guide/en/elasticsearch/reference/7.x/indices-clone-index.html>`_

        Parameters
        ----------
        index:
            The source index name
        clone_to:
            The destination index name
        override:
            If True, destination index will be removed if exists
        """
        with backend_error_handler(index=index):
            try:
                index_read_only = self.is_index_read_only(index)
                if not index_read_only:
                    self.index_read_only(index, read_only=True)
                if override:
                    self._delete_index(clone_to)
                self.__client__.indices.clone(
                    index=index,
                    target=clone_to,
                    wait_for_active_shards=settings.es_records_index_shards,
                )
            finally:
                self.index_read_only(index, read_only=index_read_only)
                self.index_read_only(clone_to, read_only=index_read_only)

    def is_index_read_only(self, index: str) -> bool:
        """
        Fetch info about read-only configuration index

        Parameters
        ----------
        index:
            The index name

        Returns
        -------
            True if queried index is read-only, False otherwise

        """
        with backend_error_handler(index=index):
            response = self.__client__.indices.get_settings(
                index=index,
                name="index.blocks.write",
                allow_no_indices=True,
                flat_settings=True,
            )
            return (
                response[index]["settings"]["index.blocks.write"] == "true"
                if response
                else False
            )

    def index_read_only(self, index: str, read_only: bool):
        """
        Enable/disable index read only

        Parameters
        ----------
        index:
            The index name
        read_only:
            True for enable read-only, False otherwise

        """
        with backend_error_handler(index=index):
            self.__client__.indices.put_settings(
                index=index,
                body={"settings": {"index.blocks.write": read_only}},
                ignore=404,
            )

    def _create_field_mapping(
        self,
        index: str,
        field_name: str,
        mapping: Dict[str, Any],
    ):
        """Creates or updates an index field mapping configuration"""
        with backend_error_handler(index=index):
            self.__client__.indices.put_mapping(
                index=index,
                body={"properties": {field_name: mapping}},
            )

    def get_cluster_info(self) -> Dict[str, Any]:
        """Returns basic about es cluster"""
        try:
            return self.__client__.info()
        except OpenSearchException as ex:
            return {"error": ex}

    def aggregate(self, index: str, aggregation: Dict[str, Any]) -> Dict[str, Any]:
        """Apply an aggregation over the index returning ONLY the agg results"""
        with backend_error_handler(index=index):
            aggregation_name = "aggregation"
            es_query = {"aggs": {aggregation_name: aggregation}}

            results = self._search(index=index, size=0, query=es_query)
            aggs_results = results["aggregations"]
            return query_helpers.parse_aggregations(aggs_results).get(aggregation_name)

    def find_metric_by_id(self, metric_id: str) -> Optional[ElasticsearchMetric]:
        metric = self.__defined_metrics__.get(metric_id)
        if not metric:
            raise EntityNotFoundError(name=metric_id, type=ElasticsearchMetric)
        return metric

    def get_task_mapping(self, task: TaskType) -> Dict[str, Any]:
        return self.__tasks_mappings__[task]

    def compute_metric(
        self,
        id: str,
        metric_id: str,
        query: Optional[Any] = None,
        params: Optional[Dict[str, Any]] = None,
    ):
        index = dataset_records_index(id)
        return self._compute_metric(
            index=index,
            metric_id=metric_id,
            query=query,
            schema=self.get_schema(id),
            params=params,
        )

    def get_schema(self, id: str) -> Dict[str, Any]:
        index = dataset_records_index(id)
        with backend_error_handler(index=index):
            response = self.__client__.indices.get_mapping(index=index)
            return self._extract_index_schema_from_response(
                index=index, es_response=response
            )

    async def update_records_content(
        self,
        id: str,
        content: Dict[str, Any],
        query: Optional[BaseDatasetsQuery],
    ) -> Tuple[int, int]:
        index = dataset_records_index(id)
        with backend_error_handler(index=index):
            es_query = self.query_builder.map_2_es_query(
                schema=self.get_schema(id),
                query=query,
            )
            response = self.client.update_by_query(
                index=index,
                body={
                    "query": es_query["query"],
                    "script": {
                        "lang": "painless",
                        "inline": ";".join(
                            [f"ctx._source.{k}='{v}'" for k, v in content.items()]
                        ),
                    },
                },
                slices="auto",
                wait_for_completion=True,
                conflicts="proceed",
            )
            total, updated = response["total"], response["updated"]
            return total, updated

    async def delete_records_by_query(
        self,
        id: str,
        query: Optional[BaseDatasetsQuery],
    ) -> Tuple[int, int]:

        index = dataset_records_index(id)
        with backend_error_handler(index=index):
            es_query = self.query_builder.map_2_es_query(
                schema=self.get_schema(id),
                query=query,
            )
            response = self.client.delete_by_query(
                index=index,
                body={"query": es_query["query"]},
                slices="auto",
                wait_for_completion=True,
                # conflicts="proceed",  # If document version conflict -> continue
            )
            total, deleted = response["total"], response["deleted"]
            return total, deleted

    def search_records(
        self,
        id: str,
        query: BackendRecordsQuery,
        sort: SortConfig,
        record_from: int = 0,
        size: int = 100,
        exclude_fields: List[str] = None,
        enable_highlight: bool = True,
    ) -> Tuple[int, List[Dict[str, Any]]]:

        index = dataset_records_index(id)
        with backend_error_handler(index=index):
            if not sort.sort_by and sort.shuffle is False:
                sort.sort_by = [SortableField(id="id")]  # Default sort by id
            es_query = {
                **self.query_builder.map_2_es_query(
                    schema=self.get_schema(id),
                    query=query,
                    sort=sort,
                ),
                "_source": {"excludes": exclude_fields or []},
                "from": record_from,
            }
            if enable_highlight:
                es_query["highlight"] = self.__configure_query_highlight__()

            results = self._search(index=index, query=es_query, size=size)
            hits = results["hits"]
            total = hits["total"]
            docs = hits["hits"]

            return total, list(map(self.__esdoc2record__, docs))

    def __esdoc2record__(
        self,
        doc: Dict[str, Any],
        is_phrase_query: bool = True,
    ):
        return {
            **doc["_source"],
            "id": doc["_id"],
            "search_keywords": self.__parse_highlight_results__(
                doc, is_phrase_query=is_phrase_query
            ),
        }

    @classmethod
    def __parse_highlight_results__(
        cls,
        doc: Dict[str, Any],
        is_phrase_query: bool = False,
    ) -> Optional[List[str]]:
        highlight_info = doc.get("highlight")
        if not highlight_info:
            return None

        search_keywords = []
        for content in highlight_info.values():
            if not isinstance(content, list):
                content = [content]
            text = " ".join(content)

            if is_phrase_query:
                text = re.sub(cls.__HIGHLIGHT_PHRASE_PRE_PARSER_REGEX__, " ", text)
            search_keywords.extend(re.findall(cls.__HIGHLIGHT_VALUES_REGEX__, text))
        return list(set(search_keywords))

    @classmethod
    def __configure_query_highlight__(cls, task: TaskType = None):

        return {
            "pre_tags": [cls.__HIGHLIGHT_PRE_TAG__],
            "post_tags": [cls.__HIGHLIGHT_POST_TAG__],
            "require_field_match": True,
            "fields": {
                "text": {},
                "text.*": {},
                "inputs.*": {},
                # **({"inputs.*": {}} if task == TaskType.text_classification else {}),
            },
        }

    # TODO(@frascuchon): Include sort parameter
    def scan_records(
        self,
        id: str,
        query: Optional[BackendRecordsQuery] = None,
        id_from: Optional[str] = None,
        limit: Optional[int] = None,
    ) -> Iterable[Dict[str, Any]]:
        index = dataset_records_index(id)
        with backend_error_handler(index):
            es_query = {
                **self.query_builder.map_2_es_query(
                    schema=self.get_schema(id),
                    query=query,
                    id_from=id_from,
                    sort=SortConfig(),  # sort by id as default for proper index scan using search after
                ),
                "highlight": self.__configure_query_highlight__(),
            }
            docs = self._list_documents(index, query=es_query, size=limit)
            for doc in docs:
                yield self.__esdoc2record__(doc)

    def open(self, id: str):
        self.open_index(dataset_records_index(id))

    def get_metadata_mappings(self, id: str):
        records_index = dataset_records_index(id)
        return self._get_field_mapping(index=records_index, field_name="metadata.*")

    def create_dataset_index(
        self,
        id: str,
        task: TaskType,
        metadata_values: Optional[Dict[str, Any]] = None,
        force_recreate: bool = False,
    ) -> None:

        _mappings = tasks_common_mappings()
        task_mappings = self.get_task_mapping(task).copy()
        for k in task_mappings:
            if isinstance(task_mappings[k], list):
                _mappings[k] = [*_mappings.get(k, []), *task_mappings[k]]
            else:
                _mappings[k] = {**_mappings.get(k, {}), **task_mappings[k]}
        index = dataset_records_index(id)
        self._create_index(
            index=index,
            settings=tasks_common_settings(),
            mappings={**tasks_common_mappings(), **_mappings},
            force_recreate=force_recreate,
        )
        if metadata_values:
            self._configure_metadata_fields(id, metadata_values)

    def _configure_metadata_fields(self, id: str, metadata_values: Dict[str, Any]):
        def check_metadata_length(metadata_length: int = 0):
            if metadata_length > settings.metadata_fields_limit:
                raise MetadataLimitExceededError(
                    length=metadata_length, limit=settings.metadata_fields_limit
                )

        def detect_nested_type(v: Any) -> bool:
            """Returns True if value match as nested value"""
            return isinstance(v, list) and isinstance(v[0], dict)

        index = dataset_records_index(id)
        check_metadata_length(len(metadata_values))
        check_metadata_length(
            len(
                {
                    *self._get_field_mapping(
                        index, "metadata.*", exclude_subfields=True
                    ),
                    *[f"metadata.{k}" for k in metadata_values.keys()],
                }
            )
        )
        for field, value in metadata_values.items():
            if detect_nested_type(value):
                self._create_field_mapping(
                    index,
                    field_name=f"metadata.{field}",
                    mapping=mappings.nested_field(),
                )

    def delete(self, id: str):
        index = dataset_records_index(id)
        try:
            self._delete_index(index, raises_error=True)
        except GenericSearchError as ex:
            if not ex.origin_error.status_code == 400:
                raise ex
            # It's an alias --> DELETE from original index
            original_index = settings.old_dataset_records_index_name.format(id)
            self._delete_index_alias(original_index, alias=index)

        finally:
            self._delete_document(index=DATASETS_INDEX_NAME, doc_id=id)

    def copy(self, id_from: str, id_to: str):
        index_from = dataset_records_index(id_from)
        index_to = dataset_records_index(id_to)

        self._clone_index(index=index_from, clone_to=index_to)

    def close(self, id: str):
        return self._close_index(dataset_records_index(id))

    def create_datasets_index(self, force_recreate: bool = False):
        self._create_index(
            DATASETS_INDEX_NAME,
            force_recreate=force_recreate,
            mappings=datasets_index_mappings(),
        )
        if not settings.enable_migration:
            return

        source_index = settings.old_dataset_index_name
        target_index = DATASETS_INDEX_NAME
        try:
            with backend_error_handler(index=source_index):
                reindex(
                    self.__client__,
                    source_index=source_index,
                    target_index=target_index,
                )
                for doc in scan(self.__client__, index=source_index):
                    dataset_id = doc["_id"]
                    index = settings.old_dataset_records_index_name.format(dataset_id)
                    alias = dataset_records_index(dataset_id)
                    self._create_index_alias(index, alias=alias)
        except IndexNotFoundError:
            pass  # Nothing to migrate

    def list_datasets(self, query: BaseDatasetsQuery):
        with backend_error_handler(index=DATASETS_INDEX_NAME):
            es_query = self.query_builder.map_2_es_query(query=query)
            return self._list_documents(
                index=DATASETS_INDEX_NAME,
                query=es_query,
                fetch_once=True,
                size=self.__MAX_NUMBER_OF_LISTED_DATASETS__,
            )

    def add_dataset_document(self, id: str, document: Dict[str, Any]):
        self._add_document(index=DATASETS_INDEX_NAME, doc_id=id, document=document)

    def update_dataset_document(self, id: str, document: Dict[str, Any]):
        self._update_document(
            index=DATASETS_INDEX_NAME,
            doc_id=id,
            document=document,
            partial_update=True,
        )

    def find_dataset(
        self, id: str, name: Optional[str] = None, owner: Optional[str] = None
    ):
        with backend_error_handler(index=DATASETS_INDEX_NAME):
            document = self._get_document_by_id(index=DATASETS_INDEX_NAME, doc_id=id)
            if not document and owner is None and name:
                # We must search by name since we have no owner
                es_query = self.query_builder.map_2_es_query(
                    query=BaseDatasetsQuery(name=name)
                )
                docs = self._list_documents(
                    index=DATASETS_INDEX_NAME,
                    query=es_query,
                    size=self.__MAX_NUMBER_OF_LISTED_DATASETS__,
                    fetch_once=True,
                )
                docs = list(docs)
                if len(docs) == 0:
                    return None

                if len(docs) > 1:
                    raise ValueError(
                        f"Ambiguous dataset info found for name {name}. Please provide a valid owner"
                    )
                document = docs[0]
            return document

    def compute_argilla_metric(self, metric_id):
        return self._compute_metric(index=DATASETS_INDEX_NAME, metric_id=metric_id)

    def _compute_metric(
        self,
        index: str,
        metric_id: str,
        query: Optional[Any] = None,
        schema: Optional[Dict[str, Any]] = None,
        params: Optional[Dict[str, Any]] = None,
    ):
        metric = self.find_metric_by_id(metric_id)
        # Only for metadata aggregation. In a future could be nice to provide the whole index schema
        params.update(
            {"schema": self._get_field_mapping(index=index, field_name="metadata.*")}
        )

        filtered_params = {
            argument: params[argument]
            for argument in metric.metric_arg_names
            if argument in params
        }

        aggs = metric.aggregation_request(**filtered_params)
        if not aggs:
            return {}
        if not isinstance(aggs, list):
            aggs = [aggs]
        results = {}
        for agg in aggs:
            es_query = {
                **self.query_builder.map_2_es_query(
                    schema=schema,
                    query=query,
                ),
                "aggs": agg,
            }
            search_result = self._search(index=index, query=es_query, size=0)
            search_aggregations = search_result.get("aggregations", {})

            if search_aggregations:
                parsed_aggregations = query_helpers.parse_aggregations(
                    search_aggregations
                )
                results.update(parsed_aggregations)

        return metric.aggregation_result(results.get(metric_id, results))

    def remove_dataset_field(self, id: str, field: str):
        self._update_document(
            index=DATASETS_INDEX_NAME,
            doc_id=id,
            script=f'ctx._source.remove("{field}")',
            partial_update=True,
        )

    def _extract_index_schema_from_response(
        self, index: str, es_response: Dict[str, Any]
    ):
        if index in es_response:
            return es_response.get(index)
        elif len(es_response) == 1:
            return list(es_response.values())[0]
        return es_response

    def _create_index_alias(self, index: str, alias: str):
        return self.__client__.indices.put_alias(
            index=index,
            name=alias,
            ignore=[400, 404],
        )

    def _delete_index_alias(self, index: str, alias: str):
        self.__client__.indices.delete_alias(index=index, name=alias, ignore=[400, 404])


def dataset_records_index(dataset_id: str) -> str:
    index_mame_template = settings.dataset_records_index_name
    return index_mame_template.format(dataset_id)



#  coding=utf-8
#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""This file reflects the user facing API.
If you want to add something here, remember to add it as normal import in the _TYPE_CHECKING section (for IDEs),
as well as in the `_import_structure` dictionary.
"""

import sys as _sys
from typing import TYPE_CHECKING as _TYPE_CHECKING

from argilla.logging import configure_logging as _configure_logging

from . import _version
from .utils import LazyargillaModule as _LazyargillaModule

__version__ = _version.version

if _TYPE_CHECKING:
    from argilla.client.api import (
        copy,
        delete,
        delete_records,
        get_workspace,
        init,
        load,
        log,
        log_async,
        set_workspace,
    )
    from argilla.client.datasets import (
        DatasetForText2Text,
        DatasetForTextClassification,
        DatasetForTokenClassification,
        read_datasets,
        read_pandas,
    )
    from argilla.client.models import (
        Text2TextRecord,
        TextClassificationRecord,
        TokenAttributions,
        TokenClassificationRecord,
    )
    from argilla.datasets import (
        TextClassificationSettings,
        TokenClassificationSettings,
        configure_dataset,
    )
    from argilla.listeners import Metrics, RBListenerContext, Search, listener
    from argilla.monitoring.model_monitor import monitor
    from argilla.server.server import app

# TODO: remove me
_import_structure = {
    "client.api": [
        "copy",
        "delete",
        "get_workspace",
        "init",
        "load",
        "delete_records",
        "log",
        "log_async",
        "set_workspace",
    ],
    "client.models": [
        "Text2TextRecord",
        "TextClassificationRecord",
        "TokenClassificationRecord",
        "TokenAttributions",
    ],
    "client.datasets": [
        "DatasetForText2Text",
        "DatasetForTextClassification",
        "DatasetForTokenClassification",
        "read_datasets",
        "read_pandas",
    ],
    "monitoring.model_monitor": ["monitor"],
    "listeners.listener": ["listener", "RBListenerContext", "Search", "Metrics"],
    "datasets": [
        "configure_dataset",
        "TextClassificationSettings",
        "TokenClassificationSettings",
    ],
    "server.app": ["app"],
}

# can be removed in a future version
_deprecated_import_structure = {
    "client.models": ["Record", "BulkResponse"],
    "client.datasets": ["Dataset"],
    "_constants": ["DEFAULT_API_KEY"],
}

_sys.modules[__name__] = _LazyargillaModule(
    __name__,
    globals()["__file__"],
    _import_structure,
    deprecated_import_structure=_deprecated_import_structure,
    module_spec=__spec__,
    extra_objects={"__version__": __version__},
)

_configure_logging()



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import copy
import dataclasses
import functools
import logging
import threading
import time
from typing import Any, Dict, List, Optional

import schedule

import argilla
from argilla.client import api
from argilla.client.sdk.commons.errors import NotFoundApiError
from argilla.listeners.models import (
    ListenerAction,
    ListenerCondition,
    Metrics,
    RBListenerContext,
    Search,
)


@dataclasses.dataclass
class RBDatasetListener:
    """
    The argilla dataset listener class

    Args:
        dataset: The dataset over which listener is created
        action: The action to execute when condition is satisfied
        metrics: A list of metrics ids that will be required in condition
        query: The query string to apply
        query_params: Defined parameters used dynamically in the provided query
        condition: The condition to satisfy to execute the action
        query_records: If ``False``, the records won't be passed as argument to the action.
            Default: ``True``
        interval_in_seconds: How often the listener is executed. Default to 30 seconds
    """

    _LOGGER = logging.getLogger(__name__)

    dataset: str
    action: ListenerAction
    metrics: Optional[List[str]] = None
    query: Optional[str] = None
    query_params: Optional[Dict[str, Any]] = None
    condition: Optional[ListenerCondition] = None
    query_records: bool = True
    interval_in_seconds: int = 30

    @property
    def formatted_query(self) -> Optional[str]:
        """Formatted query using defined query params, if any"""
        if self.query is None:
            return None
        return self.query.format(**(self.query_params or {}))

    __listener_job__: Optional[schedule.Job] = dataclasses.field(
        init=False, default=None
    )
    __stop_schedule_event__ = None
    __current_thread__ = None
    __scheduler__ = schedule.Scheduler()

    def __post_init__(self):
        self.metrics = self.metrics or []
        self._validate()

    def _validate(self):
        try:
            query = self.formatted_query
            if query:
                self._LOGGER.debug(f"Initial listener query {query}")
        except KeyError as kex:
            raise KeyError("Missing query parameter:", kex)

    def is_running(self):
        """True if listener is running"""
        return self.__listener_job__ is not None

    def __catch_exceptions__(self, cancel_on_failure=False):
        def catch_exceptions_decorator(job_func):
            @functools.wraps(job_func)
            def wrapper(*args, **kwargs):
                try:
                    return job_func(*args, **kwargs)
                except:
                    import traceback

                    print(traceback.format_exc())
                    if cancel_on_failure:
                        self.stop()  # We stop the scheduler

            return wrapper

        return catch_exceptions_decorator

    def start(self, *action_args, **action_kwargs):
        """
        Start listen to changes in the dataset. Additionally, args and kwargs can be passed to action
        by using the `action_*` arguments

        If the listener is already started, a ``ValueError`` will be raised

        """
        if self.is_running():
            raise ValueError("Listener is already running")

        job_step = self.__catch_exceptions__(cancel_on_failure=True)(
            self.__listener_iteration_job__
        )

        self.__listener_job__ = self.__scheduler__.every(
            self.interval_in_seconds
        ).seconds.do(job_step, *action_args, **action_kwargs)

        class _ScheduleThread(threading.Thread):
            _WAIT_EVENT = threading.Event()

            _THREAD_LOGGER = logging.getLogger(__name__)

            @classmethod
            def run(cls):
                cls._THREAD_LOGGER.debug("Running listener thread...")
                while not cls._WAIT_EVENT.is_set():
                    self.__scheduler__.run_pending()
                    time.sleep(self.interval_in_seconds - 1)
                cls._THREAD_LOGGER.debug("Stopping listener thread...")

            @classmethod
            def stop(cls):
                cls._WAIT_EVENT.set()

        self.__current_thread__ = _ScheduleThread()
        self.__current_thread__.start()

    def stop(self):
        """
        Stops listener if it's still running.

        If listener is already stopped, a ``ValueError`` will be raised

        """
        if not self.is_running():
            raise ValueError("Listener is not running")

        self.__scheduler__.cancel_job(self.__listener_job__)
        self.__listener_job__ = None
        self.__current_thread__.stop()
        self.__current_thread__.join()  # TODO: improve it!

    def __listener_iteration_job__(self, *args, **kwargs):
        """
        Execute a complete listener iteration. The iteration consists on:

        1. Query data and fetch configured metrics
        2. Check search results and metrics with provided condition
        3. Execute the action if condition is satisfied

        """
        current_api = api.active_api()
        try:
            dataset = current_api.datasets.find_by_name(self.dataset)
            self._LOGGER.debug(f"Found listener dataset {dataset.name}")
        except NotFoundApiError:
            self._LOGGER.warning(f"Not found dataset <{self.dataset}>")
            return

        ctx = RBListenerContext(
            listener=self,
            query_params=self.query_params,
            metrics=self.__compute_metrics__(
                current_api, dataset, query=self.formatted_query
            ),
        )
        if self.condition is None:
            self._LOGGER.debug("No condition found! Running action...")
            return self.__run_action__(ctx, *args, **kwargs)

        search_results = current_api.searches.search_records(
            name=self.dataset, task=dataset.task, query=self.formatted_query, size=0
        )

        ctx.search = Search(
            total=search_results.total, query_params=copy.deepcopy(ctx.query_params)
        )
        condition_args = [ctx.search]
        if self.metrics:
            condition_args.append(ctx.metrics)

        self._LOGGER.debug(f"Evaluate condition with arguments: {condition_args}")
        if self.condition(*condition_args):
            self._LOGGER.debug(f"Condition passed! Running action...")
            return self.__run_action__(ctx, *args, **kwargs)

    def __compute_metrics__(self, current_api, dataset, query: str) -> Metrics:
        metrics = {}
        for metric in self.metrics:
            metrics.update(
                {
                    metric: current_api.metrics.metric_summary(
                        name=self.dataset,
                        task=dataset.task,
                        metric=metric,
                        query=query,
                    )
                }
            )
        return Metrics.from_dict(metrics)

    def __run_action__(self, ctx: Optional[RBListenerContext] = None, *args, **kwargs):
        try:
            action_args = [ctx] if ctx else []
            if self.query_records:
                action_args.insert(
                    0, argilla.load(name=self.dataset, query=self.formatted_query)
                )
            self._LOGGER.debug(f"Running action with arguments: {action_args}")
            return self.action(*args, *action_args, **kwargs)
        except:
            import traceback

            print(traceback.format_exc())
            return schedule.CancelJob


def listener(
    dataset: str,
    query: Optional[str] = None,
    metrics: Optional[List[str]] = None,
    condition: Optional[ListenerCondition] = None,
    with_records: bool = True,
    execution_interval_in_seconds: int = 30,
    **query_params,
):
    """
    Configures the decorated function as a argilla listener.

    Args:
        dataset: The dataset name.
        query: The query string.
        metrics: Required metrics for listener condition.
        condition: Defines condition over search and metrics that launch action when is satisfied.
        with_records: Include records as part or action arguments. If ``False``,
            only the listener context ``RBListenerContext`` will be passed. Default: ``True``.
        execution_interval_in_seconds: Define the execution interval in seconds when listener
            iteration will be executed.
        **query_params: Dynamic parameters used in the query. These parameters will be available
            via the listener context and can be updated for subsequent queries.
    """

    def inner_decorator(func):
        return RBDatasetListener(
            dataset=dataset,
            action=func,
            condition=condition,
            query=query,
            query_params=query_params,
            metrics=metrics,
            query_records=with_records,
            interval_in_seconds=execution_interval_in_seconds,
        )

    return inner_decorator



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Optional

from pydantic import BaseModel, Field


class MetricInfo(BaseModel):
    """Metric info data model for retrieve dataset metrics information"""

    id: str = Field(description="The metric id")
    name: str = Field(description="The metric name")
    description: Optional[str] = Field(
        default=None, description="The metric description"
    )



#  coding=utf-8
#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import datetime
import json
import logging
import re
import threading
from queue import Queue
from typing import Any, Callable, Dict, List, Optional, Tuple

try:
    import starlette
except ModuleNotFoundError:
    raise ModuleNotFoundError(
        "'starlette' must be installed to use the middleware feature! "
        "You can install 'starlette' with the command: `pip install starlette>=0.13.0`"
    )
else:
    from starlette.middleware.base import BaseHTTPMiddleware, RequestResponseEndpoint
    from starlette.requests import Request
    from starlette.responses import JSONResponse, Response, StreamingResponse
    from starlette.types import Message, Receive

import argilla
from argilla.client.models import (
    Record,
    TextClassificationRecord,
    TokenClassificationRecord,
)

_logger = logging.getLogger(__name__)
_spaces_regex = re.compile(r"\s+")


def token_classification_mapper(inputs, outputs):
    if isinstance(inputs, str):
        text = inputs
    elif isinstance(inputs, dict):
        text = inputs.get("text", "")
    else:
        text = ""
    tokens = outputs.get("tokens") if isinstance(outputs, dict) else None
    return TokenClassificationRecord(
        text=text,
        tokens=tokens or _spaces_regex.split(text),
        prediction=[
            (entity["label"], entity["start"], entity["end"])
            for entity in (
                outputs.get("entities") if isinstance(outputs, dict) else outputs
            )
        ],
        event_timestamp=datetime.datetime.now(),
    )


def text_classification_mapper(inputs, outputs):
    return TextClassificationRecord(
        inputs=inputs,
        prediction=[
            (label, score)
            for label, score in zip(
                outputs.get("labels", []), outputs.get("scores", [])
            )
        ],
        event_timestamp=datetime.datetime.now(),
    )


class CachedJsonRequest(Request):
    """
    We must a cached version of incoming requests since request body cannot be read from middleware directly.
    See <https://github.com/encode/starlette/issues/847> for more information

    TODO Remove usage of CachedRequest when https://github.com/encode/starlette/pull/848 is released
    """

    @property
    def receive(self) -> Receive:
        body = None
        if hasattr(self, "_body"):
            body = self._body
        if body is not None:

            async def cached_receive() -> Message:
                return dict(type="http.request", body=body)

            return cached_receive
        return self._receive


class argillaLogHTTPMiddleware(BaseHTTPMiddleware):
    """An standard starlette middleware that enables argilla logs for http prediction requests"""

    def __init__(
        self,
        api_endpoint: str,
        dataset: str,
        records_mapper: Optional[Callable[[dict, dict], Record]] = None,
        *args,
        **kwargs
    ):
        super().__init__(*args, **kwargs)
        self._endpoint = api_endpoint
        self._dataset = dataset
        self._records_mapper = records_mapper or text_classification_mapper
        self._queue = Queue()
        self._worker_task = threading.Thread(
            target=self.__worker__, name=argillaLogHTTPMiddleware.__name__, daemon=True
        )
        self._worker_task.start()

    async def dispatch(
        self, request: Request, call_next: RequestResponseEndpoint
    ) -> Response:
        if self._endpoint != request.url.path:  # Filtering endpoint path
            return await call_next(request)

        content_type = request.headers.get("Content-type", None)
        if "application/json" not in content_type:
            return await call_next(request)

        cached_request = CachedJsonRequest(
            scope=request.scope, receive=request.receive, send=request._send
        )
        # Must read body before call_next
        inputs = await cached_request.json()
        response: Response = await call_next(cached_request)
        try:
            if (
                not isinstance(response, (JSONResponse, StreamingResponse))
                or response.status_code >= 400
            ):
                return response

            new_response, outputs = await self._extract_response_content(response)
            self._queue.put_nowait((inputs, outputs, str(request.url)))
            return new_response
        except Exception as ex:
            _logger.error("Cannot log to argilla", exc_info=ex)
            return response

    def __worker__(self):
        while True:
            try:
                inputs, outputs, url = self._queue.get()
                self._log_to_argilla(inputs, outputs, url)
            except Exception as ex:
                # Run thread FOREVER!!!
                _logger.error("Error sending records to argilla", exc_info=ex)
            finally:
                self._queue.task_done()

    async def _extract_response_content(
        self, response: Response
    ) -> Tuple[Response, List[Dict[str, Any]]]:
        """Extracts response body content from response and returns a new processable response"""
        body = b""
        new_response = response
        if isinstance(response, StreamingResponse):
            async for chunk in response.body_iterator:
                body += chunk
            new_response = StreamingResponse(
                content=(chunk for chunk in [body]),
                status_code=response.status_code,
                headers={k: v for k, v in response.headers.items()},
                media_type=response.media_type,
                background=response.background,
            )
        else:
            body = response.body
        return new_response, json.loads(body)

    def _log_to_argilla(
        self,
        inputs: List[Dict[str, Any]],
        outputs: List[Dict[str, Any]],
        url: str,
        **tags
    ):
        records = [
            record
            for _inputs, _outputs in zip(inputs, outputs)
            for record in [self._records_mapper(_inputs, _outputs)]
            if record
        ]

        if records:
            for r in records:
                r.prediction_agent = url
            argilla.log(records=records, name=self._dataset, tags=tags)



#  coding=utf-8
#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from datetime import datetime
from typing import Any, Dict, List, Optional, Type, TypeVar, cast

from fastapi import Depends

from argilla.server.daos.datasets import BaseDatasetSettingsDB, DatasetsDAO
from argilla.server.daos.models.datasets import BaseDatasetDB
from argilla.server.errors import (
    EntityAlreadyExistsError,
    EntityNotFoundError,
    ForbiddenOperationError,
    WrongTaskError,
)
from argilla.server.security.model import User


class ServiceBaseDataset(BaseDatasetDB):
    pass


class ServiceBaseDatasetSettings(BaseDatasetSettingsDB):
    pass


ServiceDataset = TypeVar("ServiceDataset", bound=ServiceBaseDataset)
ServiceDatasetSettings = TypeVar(
    "ServiceDatasetSettings", bound=ServiceBaseDatasetSettings
)


class DatasetsService:

    _INSTANCE: "DatasetsService" = None

    @classmethod
    def get_instance(
        cls, dao: DatasetsDAO = Depends(DatasetsDAO.get_instance)
    ) -> "DatasetsService":
        if not cls._INSTANCE:
            cls._INSTANCE = cls(dao)
        return cls._INSTANCE

    def __init__(self, dao: DatasetsDAO):
        self.__dao__ = dao

    def create_dataset(self, user: User, dataset: ServiceDataset) -> ServiceDataset:
        user.check_workspace(dataset.owner)

        try:
            self.find_by_name(
                user=user, name=dataset.name, task=dataset.task, workspace=dataset.owner
            )
            raise EntityAlreadyExistsError(
                name=dataset.name, type=ServiceDataset, workspace=dataset.owner
            )
        except WrongTaskError:  # Found a dataset with same name but different task
            raise EntityAlreadyExistsError(
                name=dataset.name, type=ServiceDataset, workspace=dataset.owner
            )
        except EntityNotFoundError:
            # The dataset does not exist -> create it !
            date_now = datetime.utcnow()
            dataset.created_by = user.username
            dataset.created_at = date_now
            dataset.last_updated = date_now
            return self.__dao__.create_dataset(dataset)

    def find_by_name(
        self,
        user: User,
        name: str,
        as_dataset_class: Type[ServiceDataset] = ServiceBaseDataset,
        task: Optional[str] = None,
        workspace: Optional[str] = None,
    ) -> ServiceDataset:
        owner = user.check_workspace(workspace)

        if task is None:
            found_ds = self.__find_by_name_with_superuser_fallback__(
                user, name=name, owner=owner, as_dataset_class=as_dataset_class
            )
            if found_ds:
                task = found_ds.task

        found_ds = self.__find_by_name_with_superuser_fallback__(
            user, name=name, owner=owner, task=task, as_dataset_class=as_dataset_class
        )

        if found_ds is None:
            raise EntityNotFoundError(name=name, type=ServiceDataset)
        if found_ds.owner and owner and found_ds.owner != owner:
            raise EntityNotFoundError(
                name=name, type=ServiceDataset
            ) if user.is_superuser() else ForbiddenOperationError()

        return cast(ServiceDataset, found_ds)

    def __find_by_name_with_superuser_fallback__(
        self,
        user: User,
        name: str,
        owner: Optional[str],
        as_dataset_class: Optional[Type[ServiceDataset]],
        task: Optional[str] = None,
    ):
        found_ds = self.__dao__.find_by_name(
            name=name, owner=owner, task=task, as_dataset_class=as_dataset_class
        )
        if not found_ds and user.is_superuser():
            try:
                found_ds = self.__dao__.find_by_name(
                    name=name, owner=None, task=task, as_dataset_class=as_dataset_class
                )
            except WrongTaskError:
                # A dataset exists in a different workspace and with a different task
                pass
        return found_ds

    def delete(self, user: User, dataset: ServiceDataset):
        user.check_workspace(dataset.owner)
        found = self.__find_by_name_with_superuser_fallback__(
            user=user,
            name=dataset.name,
            owner=dataset.owner,
            task=dataset.task,
            as_dataset_class=None,
        )
        if not found:
            return

        if user.is_superuser() or user.username == dataset.created_by:
            self.__dao__.delete_dataset(dataset)
        else:
            raise ForbiddenOperationError(
                f"You don't have the necessary permissions to delete this dataset. "
                "Only dataset creators or administrators can delete datasets"
            )

    def update(
        self,
        user: User,
        dataset: ServiceDataset,
        tags: Dict[str, str],
        metadata: Dict[str, Any],
    ) -> ServiceDataset:
        found = self.find_by_name(
            user=user, name=dataset.name, task=dataset.task, workspace=dataset.owner
        )

        dataset.tags = {**found.tags, **(tags or {})}
        dataset.metadata = {**found.metadata, **(metadata or {})}
        updated = found.copy(
            update={**dataset.dict(by_alias=True), "last_updated": datetime.utcnow()}
        )
        return self.__dao__.update_dataset(updated)

    def list(
        self,
        user: User,
        workspaces: Optional[List[str]],
        task2dataset_map: Dict[str, Type[ServiceDataset]] = None,
    ) -> List[ServiceDataset]:
        owners = user.check_workspaces(workspaces)
        return self.__dao__.list_datasets(
            owner_list=owners, task2dataset_map=task2dataset_map
        )

    def close(self, user: User, dataset: ServiceDataset):
        found = self.find_by_name(user=user, name=dataset.name, workspace=dataset.owner)
        self.__dao__.close(found)

    def open(self, user: User, dataset: ServiceDataset):
        found = self.find_by_name(user=user, name=dataset.name, workspace=dataset.owner)
        self.__dao__.open(found)

    def copy_dataset(
        self,
        user: User,
        dataset: ServiceDataset,
        copy_name: str,
        copy_workspace: Optional[str] = None,
        copy_tags: Dict[str, Any] = None,
        copy_metadata: Dict[str, Any] = None,
    ) -> ServiceDataset:

        dataset_workspace = copy_workspace or dataset.owner
        dataset_workspace = user.check_workspace(dataset_workspace)

        self._validate_create_dataset(
            name=copy_name,
            workspace=dataset_workspace,
            user=user,
        )

        copy_dataset = dataset.copy()
        copy_dataset.name = copy_name
        copy_dataset.owner = dataset_workspace
        date_now = datetime.utcnow()
        copy_dataset.created_at = date_now
        copy_dataset.last_updated = date_now
        copy_dataset.tags = {**copy_dataset.tags, **(copy_tags or {})}
        copy_dataset.metadata = {
            **copy_dataset.metadata,
            **(copy_metadata or {}),
            "source_workspace": dataset.owner,
            "copied_from": dataset.name,
        }

        self.__dao__.copy(
            source=dataset,
            target=copy_dataset,
        )

        return copy_dataset

    def _validate_create_dataset(self, name: str, workspace: str, user: User):
        try:
            found = self.find_by_name(user=user, name=name, workspace=workspace)
            raise EntityAlreadyExistsError(
                name=found.name,
                type=found.__class__,
                workspace=workspace,
            )
        except (EntityNotFoundError, ForbiddenOperationError):
            pass

    async def get_settings(
        self,
        user: User,
        dataset: ServiceDataset,
        class_type: Type[ServiceDatasetSettings],
    ) -> ServiceDatasetSettings:
        settings = self.__dao__.load_settings(dataset=dataset, as_class=class_type)
        if not settings:
            raise EntityNotFoundError(name=dataset.name, type=class_type)
        return class_type.parse_obj(settings.dict())

    async def save_settings(
        self, user: User, dataset: ServiceDataset, settings: ServiceDatasetSettings
    ) -> ServiceDatasetSettings:

        self.__dao__.save_settings(dataset=dataset, settings=settings)
        return settings

    async def delete_settings(self, user: User, dataset: ServiceDataset) -> None:
        self.__dao__.delete_settings(dataset=dataset)



#  coding=utf-8
#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from datetime import datetime, timedelta
from typing import Optional

from fastapi import APIRouter, Depends
from fastapi.security import (
    OAuth2PasswordBearer,
    OAuth2PasswordRequestForm,
    SecurityScopes,
)
from jose import JWTError, jwt

from argilla.server.errors import InactiveUserError, UnauthorizedError
from argilla.server.security.auth_provider.base import AuthProvider, api_key_header
from argilla.server.security.auth_provider.local.users.service import UsersService
from argilla.server.security.model import Token, User

from .settings import Settings, settings

_oauth2_scheme = OAuth2PasswordBearer(tokenUrl=settings.token_api_url, auto_error=False)


class LocalAuthProvider(AuthProvider):
    def __init__(
        self,
        users: UsersService,
        settings: Settings,
    ):
        self.users = users
        self.router = APIRouter(tags=["security"])
        self.settings = settings

        @self.router.post(
            settings.token_api_url,
            response_model=Token,
            operation_id="login_for_access_token",
        )
        async def login_for_access_token(
            form_data: OAuth2PasswordRequestForm = Depends(),
        ) -> Token:
            """
            Login access token api endpoint

            Parameters
            ----------
            form_data:
                The user/password form

            Returns
            -------
                Logging token if user is properly authenticated.
                Unauthorized exception otherwise

            """
            user = self.users.authenticate_user(form_data.username, form_data.password)
            if not user:
                raise UnauthorizedError()
            access_token_expires = timedelta(
                minutes=self.settings.token_expiration_in_minutes
            )
            access_token = self._create_access_token(
                user.username, expires_delta=access_token_expires
            )
            return Token(access_token=access_token)

    def _create_access_token(
        self, username: str, expires_delta: Optional[timedelta] = None
    ) -> str:
        """
        Creates an access token

        Parameters
        ----------
        username:
            The user name
        expires_delta:
            Token expiration

        Returns
        -------
            An access token string
        """
        to_encode = {
            "sub": username,
        }
        if expires_delta:
            to_encode["exp"] = datetime.utcnow() + expires_delta

        return jwt.encode(
            to_encode,
            self.settings.secret_key,
            algorithm=self.settings.algorithm,
        )

    def fetch_token_user(self, token: str) -> Optional[User]:
        """
        Fetch the user for a given access token

        Parameters
        ----------
        token:
            The access token

        Returns
        -------
            An User instance if a valid token was provided. None otherwise
        """
        try:
            payload = jwt.decode(
                token,
                self.settings.secret_key,
                algorithms=[self.settings.algorithm],
            )
            username: str = payload.get("sub")
            if username:
                return self.users.get_user(username=username)
        except JWTError:
            return None

    async def get_user(
        self,
        security_scopes: SecurityScopes,
        api_key: Optional[str] = Depends(api_key_header),
        token: Optional[str] = Depends(_oauth2_scheme),
    ) -> User:
        """
        Fetches the user for a given token

        Parameters
        ----------
        api_key:
            The apikey header info if provided
        token:
            The login token.
            fastapi injects this param from request
        Returns
        -------

        """
        user = await self._find_user_by_api_key(api_key)
        if user:
            return user
        if token:
            user = self.fetch_token_user(token)
        if user is None:
            raise UnauthorizedError()

        if user.disabled:
            raise InactiveUserError()

        return user

    async def _find_user_by_api_key(self, api_key) -> User:
        return await self.users.find_user_by_api_key(api_key)


def create_local_auth_provider():
    from .users.dao import create_users_dao

    settings = Settings()

    users_service = UsersService.get_instance(
        users=create_users_dao(),
    )

    return LocalAuthProvider(users=users_service, settings=settings)



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import List, Optional, Set, Type

from pydantic import BaseModel

from argilla.server.commons.models import TaskType
from argilla.server.errors import EntityNotFoundError, WrongTaskError
from argilla.server.services.datasets import ServiceDataset
from argilla.server.services.metrics import ServiceBaseMetric
from argilla.server.services.metrics.models import ServiceBaseTaskMetrics
from argilla.server.services.search.model import ServiceRecordsQuery
from argilla.server.services.tasks.commons import ServiceRecord


class TaskConfig(BaseModel):
    task: TaskType
    query: Type[ServiceRecordsQuery]
    dataset: Type[ServiceDataset]
    record: Type[ServiceRecord]
    metrics: Optional[Type[ServiceBaseTaskMetrics]]


class TasksFactory:

    __REGISTERED_TASKS__ = dict()

    @classmethod
    def register_task(
        cls,
        task_type: TaskType,
        dataset_class: Type[ServiceDataset],
        query_request: Type[ServiceRecordsQuery],
        record_class: Type[ServiceRecord],
        metrics: Optional[Type[ServiceBaseTaskMetrics]] = None,
    ):

        cls.__REGISTERED_TASKS__[task_type] = TaskConfig(
            task=task_type,
            dataset=dataset_class,
            query=query_request,
            record=record_class,
            metrics=metrics,
        )

    @classmethod
    def get_all_configs(cls) -> List[TaskConfig]:
        return [cfg for cfg in cls.__REGISTERED_TASKS__.values()]

    @classmethod
    def get_task_by_task_type(cls, task_type: TaskType) -> Optional[TaskConfig]:
        return cls.__REGISTERED_TASKS__.get(task_type)

    @classmethod
    def get_task_metrics(cls, task: TaskType) -> Optional[Type[ServiceBaseTaskMetrics]]:
        config = cls.get_task_by_task_type(task)
        if config:
            return config.metrics

    @classmethod
    def get_task_dataset(cls, task: TaskType) -> Type[ServiceDataset]:
        config = cls.__get_task_config__(task)
        return config.dataset

    @classmethod
    def get_task_record(cls, task: TaskType) -> Type[ServiceRecord]:
        config = cls.__get_task_config__(task)
        return config.record

    @classmethod
    def __get_task_config__(cls, task):
        config = cls.get_task_by_task_type(task)
        if not config:
            raise WrongTaskError(f"No configuration found for task {task}")
        return config

    @classmethod
    def find_task_metric(
        cls, task: TaskType, metric_id: str
    ) -> Optional[ServiceBaseMetric]:
        metrics = cls.find_task_metrics(task, {metric_id})
        if metrics:
            return metrics[0]
        raise EntityNotFoundError(name=metric_id, type=ServiceBaseMetric)

    @classmethod
    def find_task_metrics(
        cls, task: TaskType, metric_ids: Set[str]
    ) -> List[ServiceBaseMetric]:

        if not metric_ids:
            return []

        metrics = []
        for metric in cls.get_task_metrics(task).metrics:
            if metric.id in metric_ids:
                metrics.append(metric)
        return metrics



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import List, Optional, Tuple

from fastapi import Depends
from pydantic import BaseModel, Field

from argilla.server.daos.datasets import DatasetsDAO
from argilla.server.daos.models.records import DaoRecordsSearch
from argilla.server.daos.records import DatasetRecordsDAO
from argilla.server.errors import EntityAlreadyExistsError, EntityNotFoundError
from argilla.server.services.search.model import ServiceBaseRecordsQuery
from argilla.server.services.tasks.text_classification.model import (
    ServiceLabelingRule,
    ServiceTextClassificationDataset,
)


class DatasetLabelingRulesSummary(BaseModel):
    covered_records: int
    annotated_covered_records: int


class LabelingRuleSummary(BaseModel):
    covered_records: int
    annotated_covered_records: int
    correct_records: int = Field(default=0)
    incorrect_records: int = Field(default=0)
    precision: Optional[float] = None


class LabelingService:

    _INSTANCE = None

    @classmethod
    def get_instance(
        cls,
        datasets: DatasetsDAO = Depends(DatasetsDAO.get_instance),
        records: DatasetRecordsDAO = Depends(DatasetRecordsDAO.get_instance),
    ):
        if cls._INSTANCE is None:
            cls._INSTANCE = cls(datasets, records)
        return cls._INSTANCE

    def __init__(self, datasets: DatasetsDAO, records: DatasetRecordsDAO):
        self.__datasets__ = datasets
        self.__records__ = records

    # TODO(@frascuchon): Move all rules management methods to the common datasets service like settings
    def list_rules(
        self, dataset: ServiceTextClassificationDataset
    ) -> List[ServiceLabelingRule]:
        """List a set of rules for a given dataset"""
        return dataset.rules

    def delete_rule(self, dataset: ServiceTextClassificationDataset, rule_query: str):
        """Delete a rule from a dataset by its defined query string"""
        new_rules_set = [r for r in dataset.rules if r.query != rule_query]
        if len(dataset.rules) != new_rules_set:
            dataset.rules = new_rules_set
            self.__datasets__.update_dataset(dataset)

    def add_rule(
        self, dataset: ServiceTextClassificationDataset, rule: ServiceLabelingRule
    ) -> ServiceLabelingRule:
        """Adds a rule to a dataset"""
        for r in dataset.rules:
            if r.query == rule.query:
                raise EntityAlreadyExistsError(rule.query, type=ServiceLabelingRule)
        dataset.rules.append(rule)
        self.__datasets__.update_dataset(dataset)
        return rule

    def compute_rule_metrics(
        self,
        dataset: ServiceTextClassificationDataset,
        rule_query: str,
        labels: Optional[List[str]] = None,
    ) -> Tuple[int, int, LabelingRuleSummary]:
        """Computes metrics for given rule query and optional label against a set of rules"""

        annotated_records = self._count_annotated_records(dataset)
        dataset_records = self.__records__.search_records(dataset, size=0).total
        metric_data = self.__records__.compute_metric(
            dataset=dataset,
            metric_id="labeling_rule",
            metric_params=dict(rule_query=rule_query, labels=labels),
        )

        return (
            dataset_records,
            annotated_records,
            LabelingRuleSummary.parse_obj(metric_data),
        )

    def _count_annotated_records(
        self, dataset: ServiceTextClassificationDataset
    ) -> int:
        results = self.__records__.search_records(
            dataset,
            size=0,
            search=DaoRecordsSearch(query=ServiceBaseRecordsQuery(has_annotation=True)),
        )
        return results.total

    def all_rules_metrics(
        self, dataset: ServiceTextClassificationDataset
    ) -> Tuple[int, int, DatasetLabelingRulesSummary]:
        annotated_records = self._count_annotated_records(dataset)
        dataset_records = self.__records__.search_records(dataset, size=0).total
        metric_data = self.__records__.compute_metric(
            dataset=dataset,
            metric_id="dataset_labeling_rules",
            metric_params=dict(queries=[r.query for r in dataset.rules]),
        )

        return (
            dataset_records,
            annotated_records,
            DatasetLabelingRulesSummary.parse_obj(metric_data),
        )

    def find_rule_by_query(
        self, dataset: ServiceTextClassificationDataset, rule_query: str
    ) -> ServiceLabelingRule:
        rule_query = rule_query.strip()
        for rule in dataset.rules:
            if rule.query == rule_query:
                return rule
        raise EntityNotFoundError(rule_query, type=ServiceLabelingRule)

    def replace_rule(
        self, dataset: ServiceTextClassificationDataset, rule: ServiceLabelingRule
    ):
        for idx, r in enumerate(dataset.rules):
            if r.query == rule.query:
                dataset.rules[idx] = rule
                break
        self.__datasets__.update_dataset(dataset)



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import (
    Any,
    ClassVar,
    Dict,
    Generic,
    Iterable,
    List,
    Optional,
    TypeVar,
    Union,
)

from pydantic import BaseModel

from argilla.server.services.tasks.commons import ServiceRecord


class ServiceBaseMetric(BaseModel):
    """
    Base model for argilla dataset metrics summaries
    """

    id: str
    name: str
    description: str = None


class ServicePythonMetric(ServiceBaseMetric, Generic[ServiceRecord]):
    """
    A metric definition which will be calculated using raw queried data
    """

    def apply(self, records: Iterable[ServiceRecord]) -> Dict[str, Any]:
        """
        ServiceBaseMetric calculation method.

        Parameters
        ----------
        records:
            The matched records

        Returns
        -------
            The metric result
        """
        raise NotImplementedError()


ServiceMetric = TypeVar("ServiceMetric", bound=ServiceBaseMetric)


class ServiceBaseTaskMetrics(BaseModel):
    """
    Base class encapsulating related task metrics

    Attributes:
    -----------

    metrics:
        A list of configured metrics for task
    """

    metrics: ClassVar[List[Union[ServicePythonMetric, str]]]

    @classmethod
    def find_metric(cls, id: str) -> Optional[Union[ServicePythonMetric, str]]:
        """
        Finds a metric by id

        Parameters
        ----------
        id:
            The metric id

        Returns
        -------
            Found metric if any, ``None`` otherwise

        """
        for metric in cls.metrics:
            if isinstance(metric, str) and metric == id:
                return metric
            if metric.id == id:
                return metric

    @classmethod
    def record_metrics(cls, record: ServiceRecord) -> Dict[str, Any]:
        """
        Use this method is some configured metric requires additional
        records fields.

        Generated records will be persisted under ``metrics`` record path.
        For example, if you define a field called ``sentence_length`` like

        >>> def record_metrics(cls, record)-> Dict[str, Any]:
        ...     return { "sentence_length" : len(record.text) }

        The new field will be stored in elasticsearch in ``metrics.sentence_length``

        Parameters
        ----------
        record:
            The record used for calculate metrics fields

        Returns
        -------
            A dict with calculated metrics fields
        """
        return {}


class CommonTasksMetrics(ServiceBaseTaskMetrics, Generic[ServiceRecord]):
    """Common task metrics"""

    @classmethod
    def record_metrics(cls, record: ServiceRecord) -> Dict[str, Any]:
        """Record metrics will persist the text_length"""
        return {"text_length": len(record.all_text())}

    metrics: ClassVar[List[ServiceBaseMetric]] = [
        ServiceBaseMetric(
            id="text_length",
            name="Text length distribution",
            description="Computes the input text length distribution",
        ),
        ServiceBaseMetric(
            id="error_distribution",
            name="Error distribution",
            description="Computes the dataset error distribution. It's mean, records "
            "with correct predictions vs records with incorrect prediction "
            "vs records with unknown prediction result",
        ),
        ServiceBaseMetric(
            id="status_distribution",
            name="Record status distribution",
            description="The dataset record status distribution",
        ),
        ServiceBaseMetric(
            id="words_cloud",
            name="Inputs words cloud",
            description="The words cloud for dataset inputs",
        ),
        ServiceBaseMetric(id="metadata", name="Metadata fields stats"),
        ServiceBaseMetric(
            id="predicted_by",
            name="Predicted by distribution",
        ),
        ServiceBaseMetric(
            id="annotated_by",
            name="Annotated by distribution",
        ),
        ServiceBaseMetric(
            id="score",
            name="Score record distribution",
        ),
    ]



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from dataclasses import dataclass

from fastapi import Header, Path, Query

from argilla._constants import (
    _OLD_WORKSPACE_HEADER_NAME,
    DATASET_NAME_REGEX_PATTERN,
    WORKSPACE_HEADER_NAME,
)
from argilla.server.security.model import WORKSPACE_NAME_PATTERN

DATASET_NAME_PATH_PARAM = Path(
    ..., regex=DATASET_NAME_REGEX_PATTERN, description="The dataset name"
)


@dataclass
class RequestPagination:
    """Query pagination params"""

    limit: int = Query(50, gte=0, le=1000, description="Response records limit")
    from_: int = Query(
        0, ge=0, le=10000, alias="from", description="Record sequence from"
    )


@dataclass
class CommonTaskHandlerDependencies:
    """Common task query dependencies"""

    # TODO(@frascuchon): we could include the request user and parametrize the action scopes
    #   Depends(CommonTaskHandlerDependencies.create(scopes=[...])

    __workspace_header__: str = Header(None, alias=WORKSPACE_HEADER_NAME)
    __old_workspace_header__: str = Header(
        None,
        alias=_OLD_WORKSPACE_HEADER_NAME,
        description="This is for backward comp. with old clients",
    )
    __workspace_param__: str = Query(
        None,
        alias="workspace",
        description="The workspace where dataset belongs to. If not provided default user team will be used",
    )

    @property
    def workspace(self) -> str:
        """Return read workspace. Query param prior to header param"""
        workspace = (
            self.__workspace_param__
            or self.__workspace_header__
            or self.__old_workspace_header__
        )
        if workspace:
            assert WORKSPACE_NAME_PATTERN.match(workspace), (
                "Wrong workspace format. "
                f"Workspace must match pattern {WORKSPACE_NAME_PATTERN.pattern}"
            )
        return workspace



#  coding=utf-8
#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import asyncio
import importlib
import os
import threading
import warnings
from itertools import chain
from types import ModuleType
from typing import Any, Optional, Tuple


class LazyargillaModule(ModuleType):
    """Module class that surfaces all objects but only performs associated imports when the objects are requested.

    Shamelessly copied and adapted from the Hugging Face transformers implementation.
    """

    def __init__(
        self,
        name,
        module_file,
        import_structure,
        deprecated_import_structure=None,
        module_spec=None,
        extra_objects=None,
    ):
        super().__init__(name)
        self._modules = set(import_structure.keys())
        self._class_to_module = {}
        for key, values in import_structure.items():
            for value in values:
                self._class_to_module[value] = key
        # Needed for autocompletion in an IDE
        self.__all__ = list(import_structure.keys()) + list(
            chain(*import_structure.values())
        )
        self.__file__ = module_file
        self.__spec__ = module_spec
        self.__path__ = [os.path.dirname(module_file)]
        self._objects = {} if extra_objects is None else extra_objects
        self._name = name
        self._import_structure = import_structure

        # deprecated stuff
        deprecated_import_structure = deprecated_import_structure or {}
        self._deprecated_modules = set(deprecated_import_structure.keys())
        self._deprecated_class_to_module = {}
        for key, values in deprecated_import_structure.items():
            for value in values:
                self._deprecated_class_to_module[value] = key

    # Needed for autocompletion in an IDE
    def __dir__(self):
        result = super().__dir__()
        # The elements of self.__all__ that are submodules may or may not be in the dir already, depending on whether
        # they have been accessed or not. So we only add the elements of self.__all__ that are not already in the dir.
        for attr in self.__all__:
            if attr not in result:
                result.append(attr)
        return result

    def __getattr__(self, name: str) -> Any:
        if name in self._objects:
            return self._objects[name]
        if name in self._modules:
            value = self._get_module(name)
        elif name in self._class_to_module.keys():
            module = self._get_module(self._class_to_module[name])
            value = getattr(module, name)
        elif name in self._deprecated_modules:
            value = self._get_module(name, deprecated=True)
        elif name in self._deprecated_class_to_module.keys():
            module = self._get_module(
                self._deprecated_class_to_module[name], deprecated=True, class_name=name
            )
            value = getattr(module, name)
        else:
            raise AttributeError(f"module {self.__name__} has no attribute {name}")

        setattr(self, name, value)
        return value

    def _get_module(
        self,
        module_name: str,
        deprecated: bool = False,
        class_name: Optional[str] = None,
    ):
        if deprecated:
            warnings.warn(
                f"Importing '{class_name or module_name}' from the argilla namespace (that is "
                f"`argilla.{class_name or module_name}`) is deprecated and will not work in a future version. "
                f"Make sure you update your code accordingly.",
                category=FutureWarning,
            )

        try:
            return importlib.import_module("." + module_name, self.__name__)
        except Exception as e:
            raise RuntimeError(
                f"Failed to import {self.__name__}.{module_name} because of the following error "
                f"(look up to see its traceback):\n{e}"
            ) from e

    def __reduce__(self):
        return self.__class__, (self._name, self.__file__, self._import_structure)


def limit_value_length(data: Any, max_length: int) -> Any:
    """
    Given an input data, limits string values to a max_length by fetching
    last max_length characters

    Parameters
    ----------
    data:
        Input data
    max_length:
        Max length for string values

    Returns
    -------
        Limited version of data, if any
    """

    if isinstance(data, str):
        return data[-max_length:]
    if isinstance(data, dict):
        return {
            k: limit_value_length(v, max_length=max_length) for k, v in data.items()
        }
    if isinstance(data, (list, tuple, set)):
        new_values = map(lambda x: limit_value_length(x, max_length=max_length), data)
        return type(data)(new_values)
    return data


__LOOP__, __THREAD__ = None, None


def setup_loop_in_thread() -> Tuple[asyncio.AbstractEventLoop, threading.Thread]:
    """Sets up a new asyncio event loop in a new thread, and runs it forever.

    Returns:
        A tuple containing the event loop and the thread.
    """

    def start_background_loop(loop: asyncio.AbstractEventLoop) -> None:
        asyncio.set_event_loop(loop)
        loop.run_forever()

    global __LOOP__
    global __THREAD__

    if not (__LOOP__ and __THREAD__):
        loop = asyncio.new_event_loop()
        thread = threading.Thread(
            target=start_background_loop, args=(loop,), daemon=True
        )
        thread.start()
        __LOOP__, __THREAD__ = loop, thread
    return __LOOP__, __THREAD__



#  coding=utf-8
#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import hashlib
import logging
from enum import Enum
from typing import Dict, List, Tuple, Union

import numpy as np

from argilla import DatasetForTextClassification, TextClassificationRecord
from argilla.client.datasets import Dataset
from argilla.labeling.text_classification.weak_labels import WeakLabels, WeakMultiLabels

_LOGGER = logging.getLogger(__name__)


class TieBreakPolicy(Enum):
    """A tie break policy"""

    ABSTAIN = "abstain"
    RANDOM = "random"
    TRUE_RANDOM = "true-random"

    @classmethod
    def _missing_(cls, value):
        raise ValueError(
            f"{value} is not a valid {cls.__name__}, please select one of {list(cls._value2member_map_.keys())}"
        )


class LabelModel:
    """Abstract base class for a label model implementation.

    Args:
        weak_labels: Every label model implementation needs at least a `WeakLabels` instance.
    """

    # When we break a tie, by how much shall we increase the probability of the winner?
    _PROBABILITY_INCREASE_ON_TIE_BREAK = 0.0001

    def __init__(self, weak_labels: WeakLabels):
        self._weak_labels = weak_labels

    @property
    def weak_labels(self) -> WeakLabels:
        """The underlying `WeakLabels` object, containing the weak labels and records."""
        return self._weak_labels

    def fit(self, include_annotated_records: bool = False, *args, **kwargs):
        """Fits the label model.

        Args:
            include_annotated_records: Whether to include annotated records in the fitting.
        """
        raise NotImplementedError

    def score(self, *args, **kwargs) -> Dict:
        """Evaluates the label model."""
        raise NotImplementedError

    def predict(
        self,
        include_annotated_records: bool = False,
        prediction_agent: str = "LabelModel",
        **kwargs,
    ) -> DatasetForTextClassification:
        """Applies the label model.

        Args:
            include_annotated_records: Whether to include annotated records.
            prediction_agent: String used for the ``prediction_agent`` in the returned records.
            **kwargs: Specific to the label model implementations

        Returns:
            A dataset of records that include the predictions of the label model.
        """
        raise NotImplementedError


class MajorityVoter(LabelModel):
    """A basic label model that computes the majority vote across all rules.

    For single-label classification, it will predict the label with the most votes.
    For multi-label classification, it will predict all labels that got at least one vote by the rules.

    Args:
        weak_labels: The weak labels object.
    """

    def __init__(self, weak_labels: Union[WeakLabels, WeakMultiLabels]):
        super().__init__(weak_labels=weak_labels)

    def fit(self, *args, **kwargs):
        """Raises a NotImplementedError.

        No need to call fit on the ``MajorityVoter``!
        """
        raise NotImplementedError("No need to call fit on the 'MajorityVoter'!")

    def predict(
        self,
        include_annotated_records: bool = False,
        include_abstentions: bool = False,
        prediction_agent: str = "MajorityVoter",
        tie_break_policy: Union[TieBreakPolicy, str] = "abstain",
    ) -> DatasetForTextClassification:
        """Applies the label model.

        Args:
            include_annotated_records: Whether to include annotated records.
            include_abstentions: Whether to include records in the output, for which the label model abstained.
            prediction_agent: String used for the ``prediction_agent`` in the returned records.
            tie_break_policy: Policy to break ties (IGNORED FOR MULTI-LABEL!). You can choose among two policies:

                - `abstain`: Do not provide any prediction
                - `random`: randomly choose among tied option using deterministic hash

                The last policy can introduce quite a bit of noise, especially when the tie is among many labels,
                as is the case when all the labeling functions (rules) abstained.

        Returns:
            A dataset of records that include the predictions of the label model.
        """
        wl_matrix = self._weak_labels.matrix(
            has_annotation=None if include_annotated_records else False
        )
        records = self._weak_labels.records(
            has_annotation=None if include_annotated_records else False
        )

        if isinstance(self._weak_labels, WeakMultiLabels):
            records = self._make_multi_label_records(
                probabilities=self._compute_multi_label_probs(wl_matrix),
                records=records,
                include_abstentions=include_abstentions,
                prediction_agent=prediction_agent,
            )
        else:
            if isinstance(tie_break_policy, str):
                tie_break_policy = TieBreakPolicy(tie_break_policy)

            records = self._make_single_label_records(
                probabilities=self._compute_single_label_probs(wl_matrix),
                records=records,
                include_abstentions=include_abstentions,
                prediction_agent=prediction_agent,
                tie_break_policy=tie_break_policy,
            )

        return DatasetForTextClassification(records)

    def _compute_single_label_probs(self, wl_matrix: np.ndarray) -> np.ndarray:
        """Helper methods that computes the probabilities.

        Args:
            wl_matrix: The weak label matrix.

        Returns:
            A matrix of "probabilities" with nr or records x nr of labels.
            The label order matches the one from `self.weak_labels.labels`.
        """
        counts = np.column_stack(
            [
                np.count_nonzero(
                    wl_matrix == self._weak_labels.label2int[label], axis=1
                )
                for label in self._weak_labels.labels
            ]
        )
        with np.errstate(invalid="ignore"):
            probabilities = counts / counts.sum(axis=1).reshape(len(counts), -1)

        # we treat abstentions as ties among all labels (see snorkel)
        probabilities[np.isnan(probabilities)] = 1.0 / len(self._weak_labels.labels)

        return probabilities

    def _make_single_label_records(
        self,
        probabilities: np.ndarray,
        records: List[TextClassificationRecord],
        include_abstentions: bool,
        prediction_agent: str,
        tie_break_policy: TieBreakPolicy,
    ) -> List[TextClassificationRecord]:
        """Helper method to create records given predicted probabilities.

        Args:
            probabilities: The predicted probabilities.
            records: The records associated with the probabilities.
            include_abstentions: Whether to include records in the output, for which the label model abstained.
            prediction_agent: String used for the ``prediction_agent`` in the returned records.
            tie_break_policy: Policy to break ties. You can choose among two policies:

                - `abstain`: Do not provide any prediction
                - `random`: randomly choose among tied option using deterministic hash

                The last policy can introduce quite a bit of noise, especially when the tie is among many labels,
                as is the case when all the labeling functions (rules) abstained.

        Returns:
            A list of records that include the predictions of the label model.
        """
        records_with_prediction = []
        for i, prob, rec in zip(range(len(records)), probabilities, records):
            # Check if model abstains, that is if the highest probability is assigned to more than one label
            # 1.e-8 is taken from the abs tolerance of np.isclose
            equal_prob_idx = np.nonzero(np.abs(prob.max() - prob) < 1.0e-8)[0]
            tie = False
            if len(equal_prob_idx) > 1:
                tie = True

            # maybe skip record
            if not include_abstentions and (
                tie and tie_break_policy is TieBreakPolicy.ABSTAIN
            ):
                continue

            if not tie:
                pred_for_rec = [
                    (self._weak_labels.labels[idx], prob[idx])
                    for idx in np.argsort(prob)[::-1]
                ]
            # resolve ties following the tie break policy
            elif tie_break_policy is TieBreakPolicy.ABSTAIN:
                pred_for_rec = None
            elif tie_break_policy is TieBreakPolicy.RANDOM:
                random_idx = int(hashlib.sha1(f"{i}".encode()).hexdigest(), 16) % len(
                    equal_prob_idx
                )
                for idx in equal_prob_idx:
                    if idx == random_idx:
                        prob[idx] += self._PROBABILITY_INCREASE_ON_TIE_BREAK
                    else:
                        prob[idx] -= self._PROBABILITY_INCREASE_ON_TIE_BREAK / (
                            len(equal_prob_idx) - 1
                        )
                pred_for_rec = [
                    (self._weak_labels.labels[idx], prob[idx])
                    for idx in np.argsort(prob)[::-1]
                ]
            else:
                raise NotImplementedError(
                    f"The tie break policy '{tie_break_policy.value}' is not implemented for {self.__class__.__name__}!"
                )

            records_with_prediction.append(rec.copy(deep=True))
            records_with_prediction[-1].prediction = pred_for_rec
            records_with_prediction[-1].prediction_agent = prediction_agent

        return records_with_prediction

    def _compute_multi_label_probs(self, wl_matrix: np.ndarray) -> np.ndarray:
        """Helper methods that computes the probabilities.

        Args:
            wl_matrix: The weak label matrix.

        Returns:
            A matrix of "probabilities" with nr or records x nr of labels.
            The label order matches the one from `self.weak_labels.labels`.
        """
        # turn abstentions (-1) into 0
        counts = np.where(wl_matrix == -1, 0, wl_matrix).sum(axis=1)
        # binary probability, predict all labels with at least one vote
        probabilities = np.where(counts > 0, 1, 0).astype(np.float16)

        all_rules_abstained = wl_matrix.sum(axis=1).sum(axis=1) == (
            -1 * self._weak_labels.cardinality * len(self._weak_labels.rules)
        )
        probabilities[all_rules_abstained] = [np.nan] * len(self._weak_labels.labels)

        # more "nuanced probability", not sure if useful though
        # with np.errstate(invalid="ignore"):
        #     probabilities = counts / counts.sum(axis=1).reshape(len(counts), -1)

        return probabilities

    def _make_multi_label_records(
        self,
        probabilities: np.ndarray,
        records: List[TextClassificationRecord],
        include_abstentions: bool,
        prediction_agent: str,
    ) -> List[TextClassificationRecord]:
        """Helper method to create records given predicted probabilities.

        Args:
            probabilities: The predicted probabilities.
            records: The records associated with the probabilities.
            include_abstentions: Whether to include records in the output, for which the label model abstained.
            prediction_agent: String used for the ``prediction_agent`` in the returned records.

        Returns:
            A list of records that include the predictions of the label model.
        """
        records_with_prediction = []
        for prob, rec in zip(probabilities, records):
            all_abstained = np.isnan(prob).all()
            # maybe skip record
            if not include_abstentions and all_abstained:
                continue

            pred_for_rec = None
            if not all_abstained:
                pred_for_rec = [
                    (self._weak_labels.labels[i], prob[i])
                    for i in np.argsort(prob)[::-1]
                ]

            records_with_prediction.append(rec.copy(deep=True))
            records_with_prediction[-1].prediction = pred_for_rec
            records_with_prediction[-1].prediction_agent = prediction_agent

        return records_with_prediction

    def score(
        self,
        tie_break_policy: Union[TieBreakPolicy, str] = "abstain",
        output_str: bool = False,
    ) -> Union[Dict[str, float], str]:
        """Returns some scores/metrics of the label model with respect to the annotated records.

        The metrics are:

        - accuracy
        - micro/macro averages for precision, recall and f1
        - precision, recall, f1 and support for each label

        For more details about the metrics, check out the
        `sklearn docs <https://scikit-learn.org/stable/modules/generated/sklearn.metrics.precision_recall_fscore_support.html#sklearn-metrics-precision-recall-fscore-support>`__.

        .. note:: Metrics are only calculated over non-abstained predictions!

        Args:
            tie_break_policy: Policy to break ties (IGNORED FOR MULTI-LABEL). You can choose among two policies:

                - `abstain`: Do not provide any prediction
                - `random`: randomly choose among tied option using deterministic hash

                The last policy can introduce quite a bit of noise, especially when the tie is among many labels,
                as is the case when all the labeling functions (rules) abstained.
            output_str: If True, return output as nicely formatted string.

        Returns:
            The scores/metrics in a dictionary or as a nicely formatted str.

        Raises:
            MissingAnnotationError: If the ``weak_labels`` do not contain annotated records.
        """
        try:
            import sklearn
        except ModuleNotFoundError:
            raise ModuleNotFoundError(
                "'sklearn' must be installed to compute the metrics! "
                "You can install 'sklearn' with the command: `pip install scikit-learn`"
            )
        from sklearn.metrics import classification_report

        wl_matrix = self._weak_labels.matrix(has_annotation=True)

        if isinstance(self._weak_labels, WeakMultiLabels):
            probabilities = self._compute_multi_label_probs(wl_matrix)

            annotation, prediction = self._score_multi_label(probabilities)
            target_names = self._weak_labels.labels
        else:
            if isinstance(tie_break_policy, str):
                tie_break_policy = TieBreakPolicy(tie_break_policy)

            probabilities = self._compute_single_label_probs(wl_matrix)

            annotation, prediction = self._score_single_label(
                probabilities, tie_break_policy
            )
            target_names = self._weak_labels.labels[: annotation.max() + 1]

        return classification_report(
            annotation,
            prediction,
            target_names=target_names,
            output_dict=not output_str,
        )

    def _score_single_label(
        self, probabilities: np.ndarray, tie_break_policy: TieBreakPolicy
    ) -> Tuple[np.ndarray, np.ndarray]:
        """Helper method to compute scores for single-label classifications.

        Args:
            probabilities: The probabilities.
            tie_break_policy: Policy to break ties. You can choose among two policies:

                - `abstain`: Exclude from scores.
                - `random`: randomly choose among tied option using deterministic hash.

                The last policy can introduce quite a bit of noise, especially when the tie is among many labels,
                as is the case when all the labeling functions (rules) abstained.

        Returns:
            A tuple of the annotation and prediction array.
        """
        # 1.e-8 is taken from the abs tolerance of np.isclose
        is_max = (
            np.abs(probabilities.max(axis=1, keepdims=True) - probabilities) < 1.0e-8
        )
        is_tie = is_max.sum(axis=1) > 1

        prediction = np.argmax(is_max, axis=1)
        # we need to transform the indexes!
        annotation = np.array(
            [
                self._weak_labels.labels.index(self._weak_labels.int2label[i])
                for i in self._weak_labels.annotation()
            ],
            dtype=np.short,
        )

        if not is_tie.any():
            pass
        # resolve ties
        elif tie_break_policy is TieBreakPolicy.ABSTAIN:
            prediction, annotation = prediction[~is_tie], annotation[~is_tie]
        elif tie_break_policy is TieBreakPolicy.RANDOM:
            for i in np.nonzero(is_tie)[0]:
                equal_prob_idx = np.nonzero(is_max[i])[0]
                random_idx = int(hashlib.sha1(f"{i}".encode()).hexdigest(), 16) % len(
                    equal_prob_idx
                )
                prediction[i] = equal_prob_idx[random_idx]
        else:
            raise NotImplementedError(
                f"The tie break policy '{tie_break_policy.value}' is not implemented for MajorityVoter!"
            )

        return annotation, prediction

    def _score_multi_label(
        self, probabilities: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray]:
        """Helper method to compute scores for multi-label classifications.

        Args:
            probabilities: The probabilities.

        Returns:
            A tuple of the annotation and prediction array.
        """
        prediction = np.where(probabilities > 0.5, 1, 0)

        is_abstain = np.isnan(probabilities).all(axis=1)

        prediction, annotation = (
            prediction[~is_abstain],
            self._weak_labels.annotation()[~is_abstain],
        )

        return annotation, prediction


class Snorkel(LabelModel):
    """The label model by `Snorkel <https://github.com/snorkel-team/snorkel/>`__.

    .. note:: It is not suited for multi-label classification and does not support it!

    Args:
        weak_labels: A `WeakLabels` object containing the weak labels and records.
        verbose: Whether to show print statements
        device: What device to place the model on ('cpu' or 'cuda:0', for example).
            Passed on to the `torch.Tensor.to()` calls.

    Examples:
        >>> from argilla.labeling.text_classification import WeakLabels
        >>> weak_labels = WeakLabels(dataset="my_dataset")
        >>> label_model = Snorkel(weak_labels)
        >>> label_model.fit()
        >>> records = label_model.predict()
    """

    def __init__(
        self, weak_labels: WeakLabels, verbose: bool = True, device: str = "cpu"
    ):
        try:
            import snorkel
        except ModuleNotFoundError:
            raise ModuleNotFoundError(
                "'snorkel' must be installed to use the `Snorkel` label model! "
                "You can install 'snorkel' with the command: `pip install snorkel`"
            )
        else:
            from snorkel.labeling.model import LabelModel as SnorkelLabelModel

        super().__init__(weak_labels)

        # Check if we need to remap the weak labels to int mapping
        # Snorkel expects the abstain id to be -1 and the rest of the labels to be sequential
        if self._weak_labels.label2int[None] != -1 or sorted(
            self._weak_labels.int2label
        ) != list(range(-1, self._weak_labels.cardinality)):
            self._need_remap = True
            self._weaklabelsInt2snorkelInt = {
                self._weak_labels.label2int[label]: i
                for i, label in enumerate([None] + self._weak_labels.labels, -1)
            }
        else:
            self._need_remap = False
            self._weaklabelsInt2snorkelInt = {
                i: i for i in range(-1, self._weak_labels.cardinality)
            }

        self._snorkelInt2weaklabelsInt = {
            val: key for key, val in self._weaklabelsInt2snorkelInt.items()
        }

        # instantiate Snorkel's label model
        self._model = SnorkelLabelModel(
            cardinality=self._weak_labels.cardinality,
            verbose=verbose,
            device=device,
        )

    def fit(self, include_annotated_records: bool = False, **kwargs):
        """Fits the label model.

        Args:
            include_annotated_records: Whether to include annotated records in the fitting.
            **kwargs: Additional kwargs are passed on to Snorkel's
                `fit method <https://snorkel.readthedocs.io/en/latest/packages/_autosummary/labeling/snorkel.labeling.model.label_model.LabelModel.html#snorkel.labeling.model.label_model.LabelModel.fit>`__.
                They must not contain ``L_train``, the label matrix is provided automatically.
        """
        if "L_train" in kwargs:
            raise ValueError(
                "Your kwargs must not contain 'L_train', it is provided automatically."
            )

        l_train = self._weak_labels.matrix(
            has_annotation=None if include_annotated_records else False
        )
        if self._need_remap:
            l_train = self._copy_and_remap(l_train)

        self._model.fit(L_train=l_train, **kwargs)

    def _copy_and_remap(self, matrix_or_array: np.ndarray):
        """Helper function to copy and remap the weak label matrix or annotation array to be compatible with snorkel.

        Snorkel expects the abstain id to be -1 and the rest of the labels to be sequential.

        Args:
            matrix_or_array: The original weak label matrix or annotation array

        Returns:
            A copy of the weak label matrix, remapped to match snorkel's requirements.
        """
        matrix_or_array = matrix_or_array.copy()

        # save masks for swapping
        label_masks = {}

        # compute masks
        for idx in self._weaklabelsInt2snorkelInt:
            label_masks[idx] = matrix_or_array == idx

        # swap integers
        for idx in self._weaklabelsInt2snorkelInt:
            matrix_or_array[label_masks[idx]] = self._weaklabelsInt2snorkelInt[idx]

        return matrix_or_array

    def predict(
        self,
        include_annotated_records: bool = False,
        include_abstentions: bool = False,
        prediction_agent: str = "Snorkel",
        tie_break_policy: Union[TieBreakPolicy, str] = "abstain",
    ) -> DatasetForTextClassification:
        """Returns a list of records that contain the predictions of the label model

        Args:
            include_annotated_records: Whether to include annotated records.
            include_abstentions: Whether to include records in the output, for which the label model abstained.
            prediction_agent: String used for the ``prediction_agent`` in the returned records.
            tie_break_policy: Policy to break ties. You can choose among three policies:

                - `abstain`: Do not provide any prediction
                - `random`: randomly choose among tied option using deterministic hash
                - `true-random`: randomly choose among the tied options. NOTE: repeated runs may have slightly different results due to differences in broken ties

                The last two policies can introduce quite a bit of noise, especially when the tie is among many labels,
                as is the case when all the labeling functions (rules) abstained.

        Returns:
            A dataset of records that include the predictions of the label model.
        """
        if isinstance(tie_break_policy, str):
            tie_break_policy = TieBreakPolicy(tie_break_policy)

        l_pred = self._weak_labels.matrix(
            has_annotation=None if include_annotated_records else False
        )
        if self._need_remap:
            l_pred = self._copy_and_remap(l_pred)

        # get predictions and probabilities
        predictions, probabilities = self._model.predict(
            L=l_pred,
            return_probs=True,
            tie_break_policy=tie_break_policy.value,
        )

        # add predictions to records
        records_with_prediction = []
        for rec, pred, prob in zip(
            self._weak_labels.records(
                has_annotation=None if include_annotated_records else False
            ),
            predictions,
            probabilities,
        ):
            if not include_abstentions and pred == -1:
                continue

            records_with_prediction.append(rec.copy(deep=True))

            # set predictions to None if model abstained
            pred_for_rec = None
            if pred != -1:
                # If we have a tie, increase a bit the probability of the random winner (see tie_break_policy)
                # 1.e-8 is taken from the abs tolerance of np.isclose
                equal_prob_idx = np.nonzero(np.abs(prob.max() - prob) < 1.0e-8)[0]
                if len(equal_prob_idx) > 1:
                    for idx in equal_prob_idx:
                        if idx == pred:
                            prob[idx] += self._PROBABILITY_INCREASE_ON_TIE_BREAK
                        else:
                            prob[idx] -= self._PROBABILITY_INCREASE_ON_TIE_BREAK / (
                                len(equal_prob_idx) - 1
                            )

                pred_for_rec = [
                    (
                        self._weak_labels.int2label[
                            self._snorkelInt2weaklabelsInt[snorkel_idx]
                        ],
                        prob[snorkel_idx],
                    )
                    for snorkel_idx in np.argsort(prob)[::-1]
                ]

            records_with_prediction[-1].prediction = pred_for_rec
            records_with_prediction[-1].prediction_agent = prediction_agent

        return DatasetForTextClassification(records_with_prediction)

    def score(
        self,
        tie_break_policy: Union[TieBreakPolicy, str] = "abstain",
        output_str: bool = False,
    ) -> Union[Dict[str, float], str]:
        """Returns some scores/metrics of the label model with respect to the annotated records.

        The metrics are:

        - accuracy
        - micro/macro averages for precision, recall and f1
        - precision, recall, f1 and support for each label

        For more details about the metrics, check out the
        `sklearn docs <https://scikit-learn.org/stable/modules/generated/sklearn.metrics.precision_recall_fscore_support.html#sklearn-metrics-precision-recall-fscore-support>`__.

        .. note:: Metrics are only calculated over non-abstained predictions!

        Args:
            tie_break_policy: Policy to break ties. You can choose among three policies:

                - `abstain`: Do not provide any prediction
                - `random`: randomly choose among tied option using deterministic hash
                - `true-random`: randomly choose among the tied options. NOTE: repeated runs may have slightly different results due to differences in broken ties

                The last two policies can introduce quite a bit of noise, especially when the tie is among many labels,
                as is the case when all the labeling functions (rules) abstained.
            output_str: If True, return output as nicely formatted string.

        Returns:
            The scores/metrics in a dictionary or as a nicely formatted str.

        Raises:
            MissingAnnotationError: If the ``weak_labels`` do not contain annotated records.
        """
        from sklearn.metrics import classification_report

        if isinstance(tie_break_policy, str):
            tie_break_policy = TieBreakPolicy(tie_break_policy)

        if self._weak_labels.annotation().size == 0:
            raise MissingAnnotationError(
                "You need annotated records to compute scores/metrics for your label model."
            )

        l_pred = self._weak_labels.matrix(has_annotation=True)
        if self._need_remap:
            l_pred = self._copy_and_remap(l_pred)

        # get predictions and probabilities
        predictions, probabilities = self._model.predict(
            L=l_pred,
            return_probs=True,
            tie_break_policy=tie_break_policy.value,
        )

        # metrics are only calculated for non-abstained data points
        idx = predictions != -1

        annotation = self._weak_labels.annotation()[idx]
        if self._need_remap:
            annotation = self._copy_and_remap(annotation)

        return classification_report(
            annotation,
            predictions[idx],
            target_names=self._weak_labels.labels[: annotation.max() + 1],
            output_dict=not output_str,
        )


class FlyingSquid(LabelModel):
    """The label model by `FlyingSquid <https://github.com/HazyResearch/flyingsquid>`__.

    .. note:: It is not suited for multi-label classification and does not support it!

    Args:
        weak_labels: A `WeakLabels` object containing the weak labels and records.
        **kwargs: Passed on to the init of the FlyingSquid's
            `LabelModel <https://github.com/HazyResearch/flyingsquid/blob/master/flyingsquid/label_model.py#L18>`__.

    Examples:
        >>> from argilla.labeling.text_classification import WeakLabels
        >>> weak_labels = WeakLabels(dataset="my_dataset")
        >>> label_model = FlyingSquid(weak_labels)
        >>> label_model.fit()
        >>> records = label_model.predict()
    """

    def __init__(self, weak_labels: WeakLabels, **kwargs):
        try:
            import flyingsquid
            import pgmpy
        except ModuleNotFoundError:
            raise ModuleNotFoundError(
                "'flyingsquid' must be installed to use the `FlyingSquid` label model! "
                "You can install 'flyingsquid' with the command: `pip install pgmpy flyingsquid`"
            )
        else:
            from flyingsquid.label_model import LabelModel as FlyingSquidLabelModel

            self._FlyingSquidLabelModel = FlyingSquidLabelModel

        super().__init__(weak_labels)

        if len(self._weak_labels.rules) < 3:
            raise TooFewRulesError(
                "The FlyingSquid label model needs at least three (independent) rules!"
            )

        if "m" in kwargs:
            raise ValueError(
                "Your kwargs must not contain 'm', it is provided automatically."
            )

        self._init_kwargs = kwargs
        self._models: List[FlyingSquidLabelModel] = []

    def fit(self, include_annotated_records: bool = False, **kwargs):
        """Fits the label model.

        Args:
            include_annotated_records: Whether to include annotated records in the fitting.
            **kwargs: Passed on to the FlyingSquid's
                `LabelModel.fit() <https://github.com/HazyResearch/flyingsquid/blob/master/flyingsquid/label_model.py#L320>`__
                method.
        """
        wl_matrix = self._weak_labels.matrix(
            has_annotation=None if include_annotated_records else False
        )

        models = []
        # create a label model for each label (except for binary classification)
        # much of the implementation is taken from wrench:
        # https://github.com/JieyuZ2/wrench/blob/main/wrench/labelmodel/flyingsquid.py
        # If binary, we only need one model
        for i in range(
            1 if self._weak_labels.cardinality == 2 else self._weak_labels.cardinality
        ):
            model = self._FlyingSquidLabelModel(
                m=len(self._weak_labels.rules), **self._init_kwargs
            )
            wl_matrix_i = self._copy_and_transform_wl_matrix(wl_matrix, i)
            model.fit(L_train=wl_matrix_i, **kwargs)
            models.append(model)

        self._models = models

    def _copy_and_transform_wl_matrix(self, weak_label_matrix: np.ndarray, i: int):
        """Helper function to copy and transform the weak label matrix with respect to a target label.

         FlyingSquid expects the matrix to contain -1, 0 and 1, which are mapped the following way:

        - target label: -1
        - abstain label: 0
        - other label: 1

        Args:
            weak_label_matrix: The original weak label matrix
            i: Index of the target label

        Returns:
            A copy of the weak label matrix, transformed with respect to the target label.
        """
        wl_matrix_i = weak_label_matrix.copy()

        target_mask = (
            wl_matrix_i == self._weak_labels.label2int[self._weak_labels.labels[i]]
        )
        abstain_mask = wl_matrix_i == self._weak_labels.label2int[None]
        other_mask = (~target_mask) & (~abstain_mask)

        wl_matrix_i[target_mask] = -1
        wl_matrix_i[abstain_mask] = 0
        wl_matrix_i[other_mask] = 1

        return wl_matrix_i

    def predict(
        self,
        include_annotated_records: bool = False,
        include_abstentions: bool = False,
        prediction_agent: str = "FlyingSquid",
        verbose: bool = True,
        tie_break_policy: Union[TieBreakPolicy, str] = "abstain",
    ) -> DatasetForTextClassification:
        """Applies the label model.

        Args:
            include_annotated_records: Whether to include annotated records.
            include_abstentions: Whether to include records in the output, for which the label model abstained.
            prediction_agent: String used for the ``prediction_agent`` in the returned records.
            verbose: If True, print out messages of the progress to stderr.
            tie_break_policy: Policy to break ties. You can choose among two policies:

                - `abstain`: Do not provide any prediction
                - `random`: randomly choose among tied option using deterministic hash

                The last policy can introduce quite a bit of noise, especially when the tie is among many labels,
                as is the case when all the labeling functions (rules) abstained.

        Returns:
            A dataset of records that include the predictions of the label model.

        Raises:
            NotFittedError: If the label model was still not fitted.
        """
        if isinstance(tie_break_policy, str):
            tie_break_policy = TieBreakPolicy(tie_break_policy)

        wl_matrix = self._weak_labels.matrix(
            has_annotation=None if include_annotated_records else False
        )
        probabilities = self._predict(wl_matrix, verbose)

        # add predictions to records
        records_with_prediction = []
        for i, prob, rec in zip(
            range(len(probabilities)),
            probabilities,
            self._weak_labels.records(
                has_annotation=None if include_annotated_records else False
            ),
        ):
            # Check if model abstains, that is if the highest probability is assigned to more than one label
            # 1.e-8 is taken from the abs tolerance of np.isclose
            equal_prob_idx = np.nonzero(np.abs(prob.max() - prob) < 1.0e-8)[0]
            tie = False
            if len(equal_prob_idx) > 1:
                tie = True

            # maybe skip record
            if not include_abstentions and (
                tie and tie_break_policy is TieBreakPolicy.ABSTAIN
            ):
                continue

            if not tie:
                pred_for_rec = [
                    (self._weak_labels.labels[i], prob[i])
                    for i in np.argsort(prob)[::-1]
                ]
            # resolve ties following the tie break policy
            elif tie_break_policy is TieBreakPolicy.ABSTAIN:
                pred_for_rec = None
            elif tie_break_policy is TieBreakPolicy.RANDOM:
                random_idx = int(hashlib.sha1(f"{i}".encode()).hexdigest(), 16) % len(
                    equal_prob_idx
                )
                for idx in equal_prob_idx:
                    if idx == random_idx:
                        prob[idx] += self._PROBABILITY_INCREASE_ON_TIE_BREAK
                    else:
                        prob[idx] -= self._PROBABILITY_INCREASE_ON_TIE_BREAK / (
                            len(equal_prob_idx) - 1
                        )
                pred_for_rec = [
                    (self._weak_labels.labels[i], prob[i])
                    for i in np.argsort(prob)[::-1]
                ]
            else:
                raise NotImplementedError(
                    f"The tie break policy '{tie_break_policy.value}' is not implemented for FlyingSquid!"
                )

            records_with_prediction.append(rec.copy(deep=True))
            records_with_prediction[-1].prediction = pred_for_rec
            records_with_prediction[-1].prediction_agent = prediction_agent

        return DatasetForTextClassification(records_with_prediction)

    def _predict(self, weak_label_matrix: np.ndarray, verbose: bool) -> np.ndarray:
        """Helper function that calls the ``predict_proba`` method of FlyingSquid's label model.

        Much of the implementation is taken from wrench:
        https://github.com/JieyuZ2/wrench/blob/main/wrench/labelmodel/flyingsquid.py

        Args:
            weak_label_matrix: The weak label matrix.
            verbose: If True, print out messages of the progress to stderr.

        Returns:
            A matrix containing the probability for each label and record.

        Raises:
            NotFittedError: If the label model was still not fitted.
        """
        if not self._models:
            raise NotFittedError(
                "This FlyingSquid instance is not fitted yet. Call `fit` before using this model."
            )
        # create predictions for each label
        if self._weak_labels.cardinality > 2:
            probas = np.zeros((len(weak_label_matrix), self._weak_labels.cardinality))
            for i in range(self._weak_labels.cardinality):
                wl_matrix_i = self._copy_and_transform_wl_matrix(weak_label_matrix, i)
                probas[:, i] = self._models[i].predict_proba(
                    L_matrix=wl_matrix_i, verbose=verbose
                )[:, 0]
            probas = np.nan_to_num(probas, nan=-np.inf)  # handle NaN
            probas = np.exp(probas) / np.sum(np.exp(probas), axis=1, keepdims=True)
        # if binary, we only have one model
        else:
            wl_matrix_i = self._copy_and_transform_wl_matrix(weak_label_matrix, 0)
            probas = self._models[0].predict_proba(
                L_matrix=wl_matrix_i, verbose=verbose
            )

        return probas

    def score(
        self,
        tie_break_policy: Union[TieBreakPolicy, str] = "abstain",
        verbose: bool = False,
        output_str: bool = False,
    ) -> Union[Dict[str, float], str]:
        """Returns some scores/metrics of the label model with respect to the annotated records.

        The metrics are:

        - accuracy
        - micro/macro averages for precision, recall and f1
        - precision, recall, f1 and support for each label

        For more details about the metrics, check out the
        `sklearn docs <https://scikit-learn.org/stable/modules/generated/sklearn.metrics.precision_recall_fscore_support.html#sklearn-metrics-precision-recall-fscore-support>`__.

        .. note:: Metrics are only calculated over non-abstained predictions!

        Args:
            tie_break_policy: Policy to break ties. You can choose among two policies:

                - `abstain`: Do not provide any prediction
                - `random`: randomly choose among tied option using deterministic hash

                The last policy can introduce quite a bit of noise, especially when the tie is among many labels,
                as is the case when all the labeling functions (rules) abstained.
            verbose: If True, print out messages of the progress to stderr.
            output_str: If True, return output as nicely formatted string.

        Returns:
            The scores/metrics in a dictionary or as a nicely formatted str.

        Raises:
            NotFittedError: If the label model was still not fitted.
            MissingAnnotationError: If the ``weak_labels`` do not contain annotated records.
        """
        try:
            import sklearn
        except ModuleNotFoundError:
            raise ModuleNotFoundError(
                "'sklearn' must be installed to compute the metrics! "
                "You can install 'sklearn' with the command: `pip install scikit-learn`"
            )
        from sklearn.metrics import classification_report

        if isinstance(tie_break_policy, str):
            tie_break_policy = TieBreakPolicy(tie_break_policy)

        wl_matrix = self._weak_labels.matrix(has_annotation=True)
        probabilities = self._predict(wl_matrix, verbose)

        # 1.e-8 is taken from the abs tolerance of np.isclose
        is_max = (
            np.abs(probabilities.max(axis=1, keepdims=True) - probabilities) < 1.0e-8
        )
        is_tie = is_max.sum(axis=1) > 1

        prediction = np.argmax(is_max, axis=1)
        # we need to transform the indexes!
        annotation = np.array(
            [
                self._weak_labels.labels.index(self._weak_labels.int2label[i])
                for i in self._weak_labels.annotation()
            ],
            dtype=np.short,
        )

        if not is_tie.any():
            pass
        # resolve ties
        elif tie_break_policy is TieBreakPolicy.ABSTAIN:
            prediction, annotation = prediction[~is_tie], annotation[~is_tie]
        elif tie_break_policy is TieBreakPolicy.RANDOM:
            for i in np.nonzero(is_tie)[0]:
                equal_prob_idx = np.nonzero(is_max[i])[0]
                random_idx = int(hashlib.sha1(f"{i}".encode()).hexdigest(), 16) % len(
                    equal_prob_idx
                )
                prediction[i] = equal_prob_idx[random_idx]
        else:
            raise NotImplementedError(
                f"The tie break policy '{tie_break_policy.value}' is not implemented for FlyingSquid!"
            )

        return classification_report(
            annotation,
            prediction,
            target_names=self._weak_labels.labels[: annotation.max() + 1],
            output_dict=not output_str,
        )


class LabelModelError(Exception):
    pass


class MissingAnnotationError(LabelModelError):
    pass


class TooFewRulesError(LabelModelError):
    pass


class NotFittedError(LabelModelError):
    pass



#  coding=utf-8
#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from datetime import datetime
from typing import Dict, List, Optional, Union

from pydantic import BaseModel, Field

from argilla.client.models import (
    TextClassificationRecord as ClientTextClassificationRecord,
)
from argilla.client.models import TokenAttributions as ClientTokenAttributions
from argilla.client.sdk.commons.models import (
    MACHINE_NAME,
    BaseAnnotation,
    BaseRecord,
    PredictionStatus,
    ScoreRange,
    TaskStatus,
    UpdateDatasetRequest,
)


class ClassPrediction(BaseModel):
    class_label: Union[str, int] = Field(alias="class")
    score: float = Field(default=1.0, ge=0.0, le=1.0)


class TextClassificationAnnotation(BaseAnnotation):
    labels: List[ClassPrediction]


class TokenAttributions(BaseModel):
    token: str
    attributions: Dict[str, float] = Field(default_factory=dict)


class CreationTextClassificationRecord(BaseRecord[TextClassificationAnnotation]):
    inputs: Dict[str, Union[str, List[str]]]
    multi_label: bool = False
    explanation: Optional[Dict[str, List[TokenAttributions]]] = None

    @classmethod
    def from_client(cls, record: ClientTextClassificationRecord):
        prediction = None
        if record.prediction is not None:
            prediction = TextClassificationAnnotation(
                labels=[
                    ClassPrediction(**{"class": label, "score": score})
                    for label, score in record.prediction
                ],
                agent=record.prediction_agent or MACHINE_NAME,
            )

        annotation = None
        if record.annotation is not None:
            annotation_list = (
                record.annotation
                if isinstance(record.annotation, list)
                else [record.annotation]
            )
            annotation = TextClassificationAnnotation(
                labels=[
                    ClassPrediction(**{"class": label}) for label in annotation_list
                ],
                agent=record.annotation_agent or MACHINE_NAME,
            )

        return cls(
            inputs=record.inputs,
            prediction=prediction,
            annotation=annotation,
            multi_label=record.multi_label,
            status=record.status,
            explanation=record.explanation,
            id=record.id,
            metadata=record.metadata,
            event_timestamp=record.event_timestamp,
        )


class TextClassificationRecord(CreationTextClassificationRecord):
    last_updated: datetime = None
    _predicted: Optional[PredictionStatus] = Field(alias="predicted")

    def to_client(self) -> ClientTextClassificationRecord:
        """Returns the client model"""
        annotations = (
            [label.class_label for label in self.annotation.labels]
            if self.annotation
            else None
        )
        if annotations and not self.multi_label:
            annotations = annotations[0]

        return ClientTextClassificationRecord(
            id=self.id,
            event_timestamp=self.event_timestamp,
            inputs=self.inputs,
            multi_label=self.multi_label,
            status=self.status,
            metadata=self.metadata or {},
            prediction=[
                (label.class_label, label.score) for label in self.prediction.labels
            ]
            if self.prediction
            else None,
            prediction_agent=self.prediction.agent if self.prediction else None,
            annotation=annotations,
            annotation_agent=self.annotation.agent if self.annotation else None,
            explanation={
                key: [
                    ClientTokenAttributions.parse_obj(attribution)
                    for attribution in attributions
                ]
                for key, attributions in self.explanation.items()
            }
            if self.explanation
            else None,
            metrics=self.metrics or None,
            search_keywords=self.search_keywords or None,
        )


class TextClassificationBulkData(UpdateDatasetRequest):
    records: List[CreationTextClassificationRecord]


class TextClassificationQuery(BaseModel):
    ids: Optional[List[Union[str, int]]]

    query_text: str = Field(default=None)
    advanced_query_dsl: bool = False
    metadata: Optional[Dict[str, Union[str, List[str]]]] = None

    predicted_as: List[str] = Field(default_factory=list)
    annotated_as: List[str] = Field(default_factory=list)
    annotated_by: List[str] = Field(default_factory=list)
    predicted_by: List[str] = Field(default_factory=list)
    score: Optional[ScoreRange] = Field(default=None)
    status: List[TaskStatus] = Field(default_factory=list)
    predicted: Optional[PredictionStatus] = Field(default=None, nullable=True)

    uncovered_by_rules: List[str] = Field(
        default_factory=list,
        description="List of rule queries that WILL NOT cover the resulting records",
    )


class LabelingRule(BaseModel):
    """
    Adds read-only attributes to the labeling rule

    Attributes:
    -----------

    query:
        The ES query of the rule

    label: str
        The label associated with the rule

    description:
        A brief description of the rule

    author:
        Who created the rule

    created_at:
        When was the rule created

    """

    label: str = None
    labels: List[str] = Field(default_factory=list)
    query: str
    description: Optional[str] = None
    author: str
    created_at: datetime = None


class LabelingRuleMetricsSummary(BaseModel):
    """Metrics generated for a labeling rule"""

    coverage: Optional[float] = None
    coverage_annotated: Optional[float] = None
    correct: Optional[float] = None
    incorrect: Optional[float] = None
    precision: Optional[float] = None

    total_records: int
    annotated_records: int



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from starlette.responses import JSONResponse, StreamingResponse
from starlette.types import Send

from argilla.server.errors import APIErrorHandler


class StreamingResponseWithErrorHandling(StreamingResponse):
    async def stream_response(self, send: Send) -> None:
        try:
            return await super().stream_response(send)
        except Exception as ex:
            json_response: JSONResponse = (
                await APIErrorHandler.common_exception_handler(send, error=ex)
            )
            await send(
                {
                    "type": "http.response.body",
                    "body": json_response.body,
                    "more_body": False,
                }
            )
